import einops
import torch

from transformer_lens.HookedTransformerConfig import HookedTransformerConfig


def convert_coder_weights(model, cfg: HookedTransformerConfig):
    state_dict = {}

    state_dict["embed.W_E"] = model.transformer.wte.weight
    state_dict["pos_embed.W_pos"] = model.transformer.wpe.weight

    for l in range(cfg.n_layers):
        state_dict[f"blocks.{l}.ln1.w"] = model.transformer.h[l].ln_1.weight
        state_dict[f"blocks.{l}.ln1.b"] = model.transformer.h[l].ln_1.bias

        # In GPT-2, q,k,v are produced by one big linear map, whose output is
        # concat([q, k, v])
        W_KV = model.transformer.h[l].attn.kv_attn.weight  # [d_model, 2 * d_head]
        W_K, W_V = torch.tensor_split(W_KV, 2, dim=1)
        W_Q = model.transformer.h[l].attn.q_attn.weight  # [d_model, d_model]
        W_Q = einops.rearrange(W_Q, "m (i h)->i m h", i=cfg.n_heads)
        W_K = einops.repeat(W_K, "m h -> i m h", i=cfg.n_heads)
        W_V = einops.repeat(W_V, "m h -> i m h", i=cfg.n_heads)

        state_dict[f"blocks.{l}.attn.W_Q"] = W_Q
        state_dict[f"blocks.{l}.attn.W_K"] = W_K
        state_dict[f"blocks.{l}.attn.W_V"] = W_V

        b_Q = einops.rearrange(
            model.transformer.h[l].attn.q_attn.bias,
            "(index head)-> index head",
            index=cfg.n_heads,
            head=cfg.d_head,
        )
        b_KV = model.transformer.h[l].attn.kv_attn.bias  # [2 * d_head]
        b_K, b_V = torch.tensor_split(b_KV, 2, dim=0)
        b_K = einops.repeat(b_K, "head -> index head", index=cfg.n_heads)
        b_V = einops.repeat(b_V, "head -> index head", index=cfg.n_heads)
        state_dict[f"blocks.{l}.attn.b_Q"] = b_Q
        state_dict[f"blocks.{l}.attn.b_K"] = b_K
        state_dict[f"blocks.{l}.attn.b_V"] = b_V

        W_O = model.transformer.h[l].attn.c_proj.weight
        W_O = einops.rearrange(W_O, "(i h) m->i h m", i=cfg.n_heads)
        state_dict[f"blocks.{l}.attn.W_O"] = W_O
        state_dict[f"blocks.{l}.attn.b_O"] = model.transformer.h[l].attn.c_proj.bias

        state_dict[f"blocks.{l}.ln2.w"] = model.transformer.h[l].ln_2.weight
        state_dict[f"blocks.{l}.ln2.b"] = model.transformer.h[l].ln_2.bias

        W_in = model.transformer.h[l].mlp.c_fc.weight
        state_dict[f"blocks.{l}.mlp.W_in"] = W_in
        state_dict[f"blocks.{l}.mlp.b_in"] = model.transformer.h[l].mlp.c_fc.bias

        W_out = model.transformer.h[l].mlp.c_proj.weight
        state_dict[f"blocks.{l}.mlp.W_out"] = W_out
        state_dict[f"blocks.{l}.mlp.b_out"] = model.transformer.h[l].mlp.c_proj.bias
    state_dict["unembed.W_U"] = model.lm_head.weight.T

    state_dict["ln_final.w"] = model.transformer.ln_f.weight
    state_dict["ln_final.b"] = model.transformer.ln_f.bias
    return state_dict



from transformers import AutoConfig
from .data_classes import MapConfigClass

def convert_hf_mistral_config(hf_config: AutoConfig):

    elif architecture == "MistralForCausalLM":
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.rms_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_act,
            "normalization_type": "RMS",
            "positional_embedding_type": "rotary",
            "eps": hf_config.rms_norm_eps,
            "n_key_value_heads": hf_config.num_key_value_heads,
            "rotary_dim": hf_config.hidden_size // hf_config.num_attention_heads, #?
            "use_local_attn": True,
            "gated_mlp": True,
        }


from typing import Dict, Tuple, Union

import torch
import torch.nn as nn
from jaxtyping import Float

from transformer_lens.components import AbstractAttention
from transformer_lens.components.rms_norm import RMSNorm
from transformer_lens.HookedTransformerConfig import HookedTransformerConfig
from transformer_lens.utilities.attention import complex_attn_linear, simple_attn_linear


class GroupedQueryAttention(AbstractAttention):
    def __init__(
        self,
        cfg: Union[Dict, HookedTransformerConfig],
        attn_type: str = "global",
        layer_id: Union[int, None] = None,
    ):
        """Grouped Query Attention Block - see https://arxiv.org/abs/2305.13245 for details.
        Similar to regular attention, W_Q, W_K, and W_V all have shape [head_index, d_model, d_head].
        However, under the hood the key and value weights _W_K and _W_V are stored with shape [n_key_value_heads, d_model, d_head] and are expanded when the corresponding properties' getter is called.
        Similarly, during a forward pass, initially K and V are kept in shapes [batch, pos, n_key_value_heads, d_head] and will only be expanded to shapes [batch, pos, n_heads, d_head]
        using torch.repeat_interleave when the attention pattern and z-scores are calculated.

        Args:
            cfg (Union[Dict, HookedTransformerConfig]): Config
            attn_type (str, optional): "global" or "local", used by GPT-Neo. Local attention means the model can only attend back cfg.window_size tokens (here, 256). Not used by any other model at the moment. Defaults to "global".
            layer_id (int, optional): The index of the current layer. Used by the Mistal models (labelled here as stanford-gpt2) to scale down attention scores pre softmax for numerical stability reasons by 1/(layer_id+1). Defaults to None.
        """
        cfg = HookedTransformerConfig.unwrap(cfg)
        assert cfg.n_key_value_heads is not None
        super().__init__(cfg, attn_type, layer_id)
        self.repeat_kv_heads = cfg.n_heads // cfg.n_key_value_heads
        self._W_K = nn.Parameter(
            torch.empty(
                cfg.n_key_value_heads,
                self.cfg.d_model,
                self.cfg.d_head,
                dtype=cfg.dtype,
            )
        )
        self._W_V = nn.Parameter(
            torch.empty(
                cfg.n_key_value_heads,
                self.cfg.d_model,
                self.cfg.d_head,
                dtype=cfg.dtype,
            )
        )
        self._b_K = nn.Parameter(
            torch.zeros(cfg.n_key_value_heads, self.cfg.d_head, dtype=cfg.dtype)
        )
        self._b_V = nn.Parameter(
            torch.zeros(cfg.n_key_value_heads, self.cfg.d_head, dtype=cfg.dtype)
        )

    @property
    def W_K(self):
        return torch.repeat_interleave(self._W_K, dim=0, repeats=self.repeat_kv_heads)

    @W_K.setter
    def W_K(self, value):
        self._W_K = value

    @property
    def W_V(self):
        return torch.repeat_interleave(self._W_V, dim=0, repeats=self.repeat_kv_heads)

    @W_V.setter
    def W_V(self, value):
        self._W_V = value

    @property
    def b_K(self):
        return torch.repeat_interleave(self._b_K, dim=0, repeats=self.repeat_kv_heads)

    @b_K.setter
    def b_K(self, value):
        self._b_K = value

    @property
    def b_V(self):
        return torch.repeat_interleave(self._b_V, dim=0, repeats=self.repeat_kv_heads)

    @b_V.setter
    def b_V(self, value):
        self._b_V = value

    def calculate_qkv_matrices(
        self,
        query_input: Union[
            Float[torch.Tensor, "batch pos d_model"],
            Float[torch.Tensor, "batch pos head_index d_model"],
        ],
        key_input: Union[
            Float[torch.Tensor, "batch pos d_model"],
            Float[torch.Tensor, "batch pos kv_head_index d_model"],
        ],
        value_input: Union[
            Float[torch.Tensor, "batch pos d_model"],
            Float[torch.Tensor, "batch pos kv_head_index d_model"],
        ],
    ) -> Tuple[
        Float[torch.Tensor, "batch pos head_index d_head"],
        Float[torch.Tensor, "batch pos kv_head_index d_head"],
        Float[torch.Tensor, "batch pos kv_head_index d_head"],
    ]:
        """Calculate the Q, K, and V matrices for grouped query attention.
        This function uses the unexpanded weights _W_K and _W_V to calculate K and V.

        Args:
        query_input (Union[Float[torch.Tensor, "batch pos d_model"], Float[torch.Tensor, "batch pos head_index d_model"]]): The input tensor for the query projection.
        key_input (Union[Float[torch.Tensor, "batch pos d_model"], Float[torch.Tensor, "batch pos kv_head_index d_model"]]): The input tensor for the key projection. Note that is has as many head dimensions as the GPA block has key-value heads.
        value_input (Union[Float[torch.Tensor, "batch pos d_model"], Float[torch.Tensor, "batch pos kv_head_index d_model"]]): The input tensor for the value projection. Note that is has as many head dimensions as the GPA block has key-value heads.

        Returns:
        Tuple[Float[torch.Tensor, "batch pos head_index d_head"], Float[torch.Tensor, "batch pos kv_head_index d_head"], Float[torch.Tensor, "batch pos kv_head_index d_head"]]:
        A tuple containing the Q, K, and V matrices with the specified shapes.
        """
        attn_fn = (
            complex_attn_linear
            if self.cfg.use_split_qkv_input or self.cfg.use_attn_in
            else simple_attn_linear
        )

        q = self.hook_q(
            attn_fn(query_input, self.W_Q, self.b_Q)
        )  # [batch, pos, head_index, d_head]

        k = self.hook_k(
            attn_fn(key_input, self.W_K, self.b_K)
            if self.cfg.ungroup_grouped_query_attention
            else attn_fn(key_input, self._W_K, self._b_K)
        )  # [batch, pos, head_index, d_head]
        v = self.hook_v(
            attn_fn(value_input, self.W_V, self.b_V)
            if self.cfg.ungroup_grouped_query_attention
            else attn_fn(value_input, self._W_V, self._b_V)
        )  # [batch, pos, head_index, d_head]

        if self.cfg.use_qk_norm:
            assert self.q_norm is not None
            assert self.k_norm is not None
            q = self._apply_qk_norm(q, self.q_norm)
            k = self._apply_qk_norm(k, self.k_norm)

        return q, k, v

    def calculate_attention_scores(
        self,
        q: Float[torch.Tensor, "batch query_pos head_index d_head"],
        k: Float[torch.Tensor, "batch key_pos kv_head_index d_head"],
    ) -> Float[torch.Tensor, "batch head_index query_pos key_pos"]:
        """Calculate attention scores from Q and the unexpanded K matrix.
        K will be expaned from [batch, pos, n_key_value_head, d_head] to [batch, pos, n_query_heads, d_head] using torch.repeat_interleave.

        Args:
        q (Float[torch.Tensor, "batch query_pos head_index d_head"]): The Q tensor.
        k (Float[torch.Tensor, "batch key_pos kv_head_index d_head"]): The K tensor.

        Returns:
            Float[torch.Tensor, "batch head_index query_pos key_pos"]: The attention scores.
        """
        if not self.cfg.ungroup_grouped_query_attention:
            k = torch.repeat_interleave(k, dim=2, repeats=self.repeat_kv_heads)
        return super().calculate_attention_scores(q, k)

    def calculate_z_scores(
        self,
        v: Float[torch.Tensor, "batch key_pos kv_head_index d_head"],
        pattern: Float[torch.Tensor, "batch head_index query_pos key_pos"],
    ) -> Float[torch.Tensor, "batch query_pos head_index d_head"]:
        """Calculate z scores from the attention pattern and the unexpanded V matrix.
        V will be expaned from [batch, pos, n_key_value_head, d_head] to [batch, pos, n_query_heads, d_head] using torch.repeat_interleave.

        Args:
        v (Float[torch.Tensor, "batch query_pos head_index d_head"]): The V tensor.
        pattern (Float[torch.Tensor, "batch key_pos kv_head_index d_head"]): The attention pattern.

        Returns:
            Float[torch.Tensor, "batch head_index query_pos key_pos"]: The z scores.
        """
        if not self.cfg.ungroup_grouped_query_attention:
            v = torch.repeat_interleave(v, dim=2, repeats=self.repeat_kv_heads)
        return super().calculate_z_scores(v, pattern)

    def _apply_qk_norm(
        self, x: Float[torch.Tensor, "batch pos head_index d_head"], norm_module: RMSNorm
    ) -> Float[torch.Tensor, "batch pos head_index d_head"]:
        """Apply QK normalization with proper reshaping.

        Args:
            x: Input tensor with shape [batch, pos, head_index, d_head]
            norm_module: RMSNorm module to apply

        Returns:
            Normalized tensor with same shape as input
        """
        # Reshape from [batch, pos, head_index, d_head] to [batch * pos * head_index, d_head]
        batch, pos, n_heads, d_head = x.shape
        x_reshaped = x.reshape(-1, d_head)
        x_normed = norm_module(x_reshaped)
        return x_normed.reshape(batch, pos, n_heads, d_head)



"""Hooked Transformer Layer Norm Pre Component.

This module contains all the component :class:`LayerNormPre`.
"""
from typing import Dict, Union

import torch
import torch.nn as nn
from jaxtyping import Float

from transformer_lens.hook_points import HookPoint
from transformer_lens.HookedTransformerConfig import HookedTransformerConfig


# LayerNormPre
# I fold the LayerNorm weights and biases into later weights and biases.
# This is just the 'center and normalise' part of LayerNorm
# Centering is equivalent to just deleting one direction of residual space,
# and is equivalent to centering the weight matrices of everything writing to the residual stream
# Normalising is a funkier non-linear operation, that projects the residual stream onto the unit hypersphere
class LayerNormPre(nn.Module):
    def __init__(self, cfg: Union[Dict, HookedTransformerConfig]):
        """LayerNormPre - the 'center and normalise' part of LayerNorm. Length is
        normally d_model, but is d_mlp for softmax. Not needed as a parameter. This
        should only be used in inference mode after folding in LayerNorm weights"""
        super().__init__()
        self.cfg = HookedTransformerConfig.unwrap(cfg)
        self.eps = self.cfg.eps

        # Adds a hook point for the normalisation scale factor
        self.hook_scale = HookPoint()  # [batch, pos]
        # Hook Normalized captures LN output - here it's a vector with std 1 and mean 0
        self.hook_normalized = HookPoint()  # [batch, pos, length]

    def forward(
        self,
        x: Union[
            Float[torch.Tensor, "batch pos d_model"],
            Float[torch.Tensor, "batch pos head_index d_model"],
        ],
    ) -> Union[
        Float[torch.Tensor, "batch pos d_model"],
        Float[torch.Tensor, "batch pos head_index d_model"],
    ]:
        if self.cfg.dtype not in [torch.float32, torch.float64]:
            x = x.to(torch.float32)

        x = x - x.mean(-1, keepdim=True)  # [batch, pos, length]
        scale: Union[
            Float[torch.Tensor, "batch pos 1"],
            Float[torch.Tensor, "batch pos head_index 1"],
        ] = self.hook_scale((x.pow(2).mean(-1, keepdim=True) + self.eps).sqrt())
        return self.hook_normalized(x / scale).to(self.cfg.dtype)



from . import hook_points
from . import utils
from . import evals
from .past_key_value_caching import (
    HookedTransformerKeyValueCache,
    HookedTransformerKeyValueCacheEntry,
)
from . import components
from . import factories
from .HookedTransformerConfig import HookedTransformerConfig
from .FactoredMatrix import FactoredMatrix
from .ActivationCache import ActivationCache
from .HookedTransformer import HookedTransformer
from .SVDInterpreter import SVDInterpreter
from .HookedEncoder import HookedEncoder
from .HookedEncoderDecoder import HookedEncoderDecoder
from .BertNextSentencePrediction import BertNextSentencePrediction
from . import head_detector
from . import loading_from_pretrained as loading
from . import patching
from . import train

from .past_key_value_caching import (
    HookedTransformerKeyValueCache as EasyTransformerKeyValueCache,
    HookedTransformerKeyValueCacheEntry as EasyTransformerKeyValueCacheEntry,
)
from .HookedTransformer import HookedTransformer as EasyTransformer
from .HookedTransformerConfig import HookedTransformerConfig as EasyTransformerConfig



from collections import defaultdict
from typing import List, Tuple

import einops
import torch
import torch.nn as nn
from accelerate import Accelerator
from torch import Tensor as TT
from transformers import (AutoImageProcessor, AutoModelForCausalLM,
                          AutoModelForImageClassification, AutoTokenizer,
                          PreTrainedModel)

from .data_classes import DtypeMap
from .hooks import HookConfig, HookMap
from .model_maps import ConfigClass, ModelMap, convert_hf_model_config


class Model:
    def __init__(self,
            model_repo: str = "nickypro/tinyllama-15m", # Which huggingface model to load
            limit: int = None, # Max amount of tokens to allow to run
            model_device: str = None, # device for transformers model ["cuda", "cpu", "mps"]
            output_device: str = None, # device for saving outputs
            device_map: str = None, # transformers device map config ["auto", "cuda"]
            use_accelerator: bool = True, # Allow multigpu
            dtype: str = "bfp16", # See DtypeConfig. ["bfp16", "fp32", "fp16", "hqq8", "int8", "hqq4", "int4", "nf4"]
            compile: bool = False, # whether to use compiled backend. Prevents "training/backprop", adds speed
            torch_dtype: torch.dtype = None, # manual torch.dtype
            svd_attn: bool = False, # whether to modify attention weights with SVD (TODO: reimplement)
            tokenizer_repo: str = None, # huggingface tokenizer to load (defaults to model_repo)
            mask_fn: str = "step", # what kind of mask to apply to model hooks ("step", "sigmoid", ...)
            eval_mode: bool = True, # set model to model.eval(), not sure what this does
            hook_config: str = None, # See HookConfig. hook configuration string.
            add_hooks: bool = True, # whether to add hooks from the start or not [True, False]
            model_kwargs: dict = None, # additional kwargs for the model
        ):
        # Handle devices and multi-gpu stuff
        self.use_accelerator = False
        if (model_device is None) and (use_accelerator) and (torch.cuda.device_count() > 1):
            self.use_accelerator = True
            self.accelerator = Accelerator()
            self.device = self.accelerator.device
        else:
            self.device = model_device or ("cuda" if torch.cuda.is_available() else "cpu")

        # Model device mapping and output device
        self.device_map = device_map or "auto"
        if not self.use_accelerator and self.device != "cuda":
            self.device_map = self.device
        self.output_device = output_device or self.device

        # Handle dtype
        self.dtype_map = DtypeMap(dtype, torch_dtype)
        self.dtype = self.dtype_map._dtype
        self.dtype_args = self.dtype_map._dtype_args
        self.use_quantization = "quantization_config" in self.dtype_args

        # Model configuration
        self.model_repo = model_repo
        self.tokenizer_repo = tokenizer_repo or model_repo
        self.svd_attn = svd_attn
        self.mask_fn = mask_fn
        self.eval_mode = eval_mode
        self.limit = limit

        # Initialize model components
        self.cfg: ConfigClass = None
        self.tokenizer: AutoTokenizer = None
        self.processor: AutoImageProcessor = None
        self.predictor: AutoModelForCausalLM | AutoModelForImageClassification = None
        self.orig_predictor: AutoModelForCausalLM | AutoModelForImageClassification = None
        self.peft_predictor = None
        self.map: ModelMap = None
        self.model: PreTrainedModel = None
        self.layers: list = None

        # Hooking into the model
        self.hook_config: HookConfig = hook_config or self.default_config()
        self.hooks: HookMap = HookMap(self.hook_config)
        self.hook_config = self.hooks.hook_config

        # Initialize the model
        self.compile = compile
        self.model_kwargs = model_kwargs or {}
        self.init_model(add_hooks=add_hooks)
        self.run_example_input()

    @staticmethod
    def default_config():
        config_string = """
        pre_decoder: collect
        post_decoder: collect
        pre_attn: collect
        attn_pre_out: offset, mask, replace, collect, unoffset
        post_attn: collect
        pre_mlp: collect
        mlp_pre_out: offset, mask, replace, collect, unoffset
        post_mlp: collect
        """
        return HookConfig().from_string(config_string)

    def show_details( self, verbose=True ):
        if verbose:
            print(" - n_layers :", self.cfg.n_layers)
            print(" - d_model  :", self.cfg.d_model)
            print(" - n_heads  :", self.cfg.n_heads)
            print(" - d_head   :", self.cfg.d_head)
            print(" - d_mlp    :", self.cfg.d_mlp)
        else:
            print( f" - n_layers, d_model = {self.cfg.n_layers}, {self.cfg.d_model}" )

    def import_models(self,
            tokenizer=None,
            predictor=None,
            processor=None
        ):
        # Import model components (Default: Causal Language Models)
        device_map = self.device_map
        model_args = {**self.dtype_args, **self.model_kwargs}

        if self.cfg.model_modality == "vision":
            self.tokenizer = None \
                if tokenizer is None else tokenizer
            self.processor = self.init_image_processor(device_map) \
                if processor is None else self.processor
            self.predictor = AutoModelForImageClassification.from_pretrained(
                self.model_repo, device_map=device_map, **model_args) \
                if predictor is None else predictor
        elif self.cfg.model_modality == "language":
            self.tokenizer = AutoTokenizer.from_pretrained(self.tokenizer_repo, legacy=False, padding_side='left') \
                if tokenizer is None else tokenizer
            self.processor = None \
                if processor is None else processor
            self.predictor = AutoModelForCausalLM.from_pretrained(
                self.model_repo, device_map=device_map, **model_args) \
                if predictor is None else predictor

        else:
            raise NotImplementedError(f"Model modality {self.cfg.model_modality} not implemented.")
        self.orig_predictor = self.predictor
        print(f"Loaded model '{self.model_repo}' with {self.dtype_map.str_dtype}:")

    def init_model(self, do_model_import=True, add_hooks=True):
        self.cfg = convert_hf_model_config(self.model_repo)
        if do_model_import:
            self.import_models()
        self.map = ModelMap(self.predictor, self.cfg)
        self.model = self.map["model"] #.to(self.device)
        self.layers = self.map.layers
        if add_hooks:
            self.init_hooks()
        if self.compile:
            self.init_compile()

    def init_peft(self, peft_config_or_path, reinit_hooks=False):
        """
        Initialize PEFT for the model.

        Args:
            peft_config_or_path (PeftConfig or str):
                Either a PeftConfig object or a string path to a PEFT model repository.
        """
        try:
            from peft import PeftConfig, PeftModel, get_peft_model
        except ImportError:
            raise ImportError("PEFT is not installed. Please install it with 'pip install peft'.")

        if reinit_hooks:
            self.remove_hooks()
        if isinstance(peft_config_or_path, str): # load from repo
            self.peft_predictor = PeftModel.from_pretrained(self.orig_predictor, peft_config_or_path)
        else: # create from config
            self.peft_predictor = get_peft_model(self.orig_predictor, peft_config_or_path)
        self.predictor = self.peft_predictor.base_model.model
        # Reinitialize the model
        self.init_model(do_model_import=False, add_hooks=False)
        print(f"Initialized PEFT model")
        self.peft_predictor.print_trainable_parameters()
        if reinit_hooks:
            self.init_hooks()

    def init_compile(self):
        from torch import _dynamo
        torch._dynamo.config.suppress_errors = True
        self.predictor = self.dtype_map.compile(self.predictor)

    def __getitem__(self, key):
        return self.map[key]

    def __setitem__(self, key, value):
        self.map[key] = value

    def run_example_input(self):
        with torch.no_grad():
            if self.cfg.model_modality == "language":
                self.get_outputs_embeds("a b")
            if self.cfg.model_modality == "vision":
                self.get_outputs_embeds(
                    pixel_values=torch.randn([1,3,self.cfg.image_size,self.cfg.image_size], dtype=self.dtype, device=self.device)
                )

    def init_image_processor(self, device_map):
        """ Initialize processor from raw pixel values to normalised tensors"""
        try:
            from transformers import AutoImageProcessor
            self.processor = AutoImageProcessor.from_pretrained(
                self.model_repo, device_map=device_map, **self.dtype_args)
        except:
            from .vit_processor import SsdVitProcessor
            self.processor = SsdVitProcessor()
        return self.processor

    # Functions for hooks
    def set_hook_config(self, config: HookConfig):
        self.remove_hooks()
        if isinstance(config, str):
            config = HookConfig().from_string(config)
        self.hook_config = config
        self.hooks = HookMap(config)
        self.init_hooks()
        self.run_example_input()

    def init_hooks(self):
        # (x) [attn_out] (y) -> (x) [layer] (y)
        num_hooks: int = 0
        for layer_idx, layer in enumerate(self.layers):
            for point in self.hook_config.hook_points.keys():
                hooks = self.hook_config.get_hooks(point, layer_idx)
                if hooks:
                    io_type, module = self.get_module_for_hook_point(layer, point)
                    self.register_hooks(f"layer_{layer_idx}_{point}", module, hooks, io_type)
                    num_hooks += len(hooks)
        self.hook_config.n_layers = self.cfg.n_layers
        print(f"- Added {num_hooks} hooks across {layer_idx+1} layers")

    def register_hooks(self, name, module, hooks, io_type):
        hook_fn = self.get_hook_fn(name, hooks, io_type)
        if io_type == "in":
            handle = module.register_forward_pre_hook(hook_fn)
        elif io_type == "out":
            handle = module.register_forward_hook(hook_fn)
        self.hooks.handles.append(handle)

    def get_hook_fn(self, name, hooks, io_type="in"):
        def hook_fn(module, __input, __output=None):
            # Is this input or output?
            __act = __input if io_type=="in" else __output
            activation = __act[0] if isinstance(__act, tuple) else __act

            # Does the module not use batch index? if so, we fix this
            has_batch = f"has_batch_index_{io_type}"
            if not hasattr(module, has_batch): # init run has 1 batch only
                setattr(module, has_batch, activation.shape[0] == 1)
            if not getattr(module, has_batch):
                activation = activation.unsqueeze(dim=0)

            if "attn_pre_out" in name:
                activation = self.split_attn_head_dims(activation)

            for hook in hooks:
                # split attention to [n_heads, d_head]
                curr_hook  = self.hooks.get_hook_fn(hook, name, activation, self.device, self.dtype)
                activation = curr_hook(activation)

            if "attn_pre_out" in name:
                activation = self.join_attn_head_dims(activation)
            return (activation,) + __act[1:] if isinstance(__act, tuple) else activation

        return hook_fn

    # Hook-specific methods
    def get_module_for_hook_point(self, layer, point):
        layer["attn"].is_attention = True
        # Attention Points
        if point == "pre_attn":
            return "in", layer["attn.ln_in"] if self.cfg.pre_layernorm else layer["attn"]
        elif point == "attn_pre_out":
            return "in", layer["attn.out_proj"]
        elif point == "post_attn":
            return "out", layer["attn.ln_out"] if self.cfg.post_layernorm else layer["attn"]
        # MLP Points
        elif point == "pre_mlp":
            return "in", layer["mlp.ln_in"] if self.cfg.pre_layernorm else layer["mlp.in_proj"]
        elif point == "mlp_pre_out":
            return "in", layer["mlp.out_proj"]
        elif point == "post_mlp":
            return "out", layer["mlp.ln_out"] if self.cfg.post_layernorm else layer["mlp.out_proj"]
        # decoder block points:
        elif point == "pre_decoder":
            return "in", layer.layer
        elif point == "post_decoder":
            return "out", layer.layer
        else:
            raise ValueError(f"Unknown hook point: {point}")

    def remove_hooks(self):
        for handle in self.hooks.handles:
            handle.remove()
        self.hooks.handles.clear()


    # Helper functions for reshaping activations
    def split_attn_head_dims(self, activation):
        if activation.shape[-1] == (self.cfg.n_heads * self.cfg.d_head):
            activation = einops.rearrange(
                activation, "... (n_heads d_head) -> ... n_heads d_head",
                n_heads=self.cfg.n_heads, d_head=self.cfg.d_head)
        return activation

    def join_attn_head_dims(self, activation):
        if activation.shape[-1] == self.cfg.d_head:
            activation = einops.rearrange(
                activation, "... n_heads d_head -> ... (n_heads d_head)",
                n_heads=self.cfg.n_heads, d_head=self.cfg.d_head)
        return activation

    # Functions for symmetrically altering model.
    def center_unembed(self):
        with torch.no_grad():
            if "unembed" in self.map:
                lm_head = self.map["unembed"]
            else:
                lm_head = self.model.get_output_embeddings()
            embed = self.map["embed"]
            assert embed is not lm_head, "centering unembedding not yet supported for tied weights"
            lm_head.weight = torch.nn.Parameter(lm_head.weight - lm_head.weight.mean(dim=-1, keepdim=True))

    # Get "normal" Model Activations
    # text           --[tokenizer]-----> input_ids
    def get_ids(self, text:str):
        input_ids = self.tokenizer( text, return_tensors='pt').input_ids
        if self.limit is not None:
            input_ids = torch.stack([ input_ids[0][:self.limit] ])
        return input_ids.to( self.device )


    # raw_img      --[processor]----> pixel_values
    def get_pixel_values(self, raw_img):
        img = self.processor(raw_img, return_tensors="pt")
        pixel_values = img["pixel_values"].to(self.device)
        return pixel_values

    # input_ids    --[model.embed]--> inputs_embeds
    # pixel_values --[model.embed]--> inputs_embeds
    def get_inputs_embeds(self, text:str = None, input_ids:TT = None, raw_img=None, pixel_values=None):
        if text is not None:
            input_ids = self.get_ids(text)
        if input_ids is not None:
            return self.map["embed"](input_ids)
        if raw_img is not None:
            pixel_values = self.get_pixel_values(raw_img)
        if pixel_values is not None:
            return self.map["embed"](pixel_values)
        return None

    # inputs_embeds  --[model.model]---> outputs_embeds
    def get_outputs_embeds(self, text:str=None, input_ids:TT=None, raw_img=None, pixel_values=None, inputs_embeds:TT=None, attention_mask=None, **kwargs):
        """Get output logits from input text/image"""
        if attention_mask is not None: # in order for attention to work right need to do this ???
            kwargs["output_attentions"] = True
        if self.cfg.model_modality == "vision":
            pixel_values = self.get_pixel_values(raw_img) if pixel_values is None else pixel_values
            return self.model(pixel_values, output_hidden_states=False, **kwargs ).last_hidden_state

        inputs_embeds = self.get_inputs_embeds(text, input_ids) \
            if inputs_embeds is None else inputs_embeds
        return self.model( inputs_embeds=inputs_embeds, output_hidden_states=False, attention_mask=attention_mask, **kwargs ).last_hidden_state

    def get_attn_weights(self, text=None, input_ids=None, inputs_embeds=None):
        inputs_embeds = inputs_embeds if inputs_embeds is not None \
            else self.get_inputs_embeds(text=text, input_ids=input_ids)
        outputs = self.model(inputs_embeds=inputs_embeds, output_attentions=True)
        return einops.rearrange(torch.stack(outputs.attentions),
            "layer batch head tok_i tok_j -> batch layer head tok_i tok_j",
            layer=self.cfg.n_layers, head=self.cfg.n_heads,
        )

    # outputs_embeds --[model.lm_head]-> logits
    def unembed(self, embedded_outputs: TT):
        """ Converts outputs_embeds -> token logits. Is also basically LogitLens."""
        if "unembed" in self.map.key_map: #aka "lm_head"
            lm_head = self.map["unembed"]
        else:
            lm_head = self.predictor.get_output_embeddings()
        return lm_head( embedded_outputs.to(self.device) )

    def get_logits(self, text:str=None, input_ids:TT=None, raw_img=None, pixel_values=None, inputs_embeds:TT=None, outputs_embeds:TT=None, attention_mask=None, **kwargs):
        outputs_embeds = self.get_outputs_embeds(text, input_ids, raw_img, pixel_values, inputs_embeds, attention_mask, **kwargs) \
            if outputs_embeds is None else outputs_embeds
        logits = self.unembed(outputs_embeds)
        return logits


    def generate(self, text:str=None, num=10, max_length=None,
            input_ids: TT = None,
            inputs_embeds: TT = None,
            do_sample: bool = True,
            temperature: float = 0.7,
            **kwargs,
        ):
        """ Predict the next {num} tokens from an input {text}."""

        if text is not None:
            input_ids = self.get_ids(text).to(self.device)

        if input_ids is not None:
            inputs_embeds = self.get_inputs_embeds(input_ids=input_ids).to(self.device)

        if input_ids is None:
            text_before = ""
            input_ids = torch.zeros(inputs_embeds.shape[:-1], dtype=torch.long).to(self.device)
        else:
            text_before = self.tokenizer.batch_decode( input_ids,
                skip_special_tokens=True, clean_up_tokenization_spaces=False )[0]

        assert inputs_embeds is not None, "must pass text, input_ids, or inputs_embeds"

        # Write an attention mask so that the tokenizer doesn't complain
        attn_mask = None
        if hasattr(self.tokenizer, "pad_token_id"):
            attn_mask = torch.ones_like(input_ids).bool()
            for index, _id in enumerate(attn_mask[0]):
                if _id == self.tokenizer.pad_token_id:
                    attn_mask[index] = 0

        # Hard code GPT2 Tokeniser pad_token_id to avoid warnings
        if self.cfg.architecture == "GPT2LMHeadModel":
            if "pad_token_id" not in kwargs:
                kwargs["pad_token_id"] = 50256

        new_length = len(input_ids[0])+num if max_length is None else max_length

        # Generate from input_ids if available, seems to work more reliably (eg gemma2)
        if input_ids is not None:
            generate_ids = self.predictor.generate(input_ids=input_ids, max_length=new_length,
                do_sample=do_sample, temperature=temperature, pad_token_id=self.tokenizer.pad_token_ids,
                attention_mask=attn_mask, **kwargs)
            text_after = self.tokenizer.batch_decode( generate_ids[:, len(input_ids[0]):],
                skip_special_tokens=True, clean_up_tokenization_spaces=False )[0]
            return text_before, text_after

        # Otherwise, generate from inputs_embeds
        generate_ids = self.predictor.generate(inputs_embeds=inputs_embeds, max_length=new_length,
            do_sample=do_sample, temperature=temperature,
            attention_mask=attn_mask, **kwargs)
        text_after  = self.tokenizer.batch_decode( generate_ids,
            skip_special_tokens=True, clean_up_tokenization_spaces=False )[0]
        return text_before, text_after

    def generate_batch(self,
            batch_prompts: List[str],
            num: int = 10,
            max_length: int = None,
            temperature: float = 0.7,
            do_sample: bool = True,
            **kwargs
        ) -> Tuple[List[str], List[str]]:
        """
        Generate the next {num} tokens for a batch of input prompts.

        Args:
            batch_prompts (List[str]): A list of input prompts.
            num (int, optional): Number of new tokens to generate per prompt. Defaults to 10.
            max_length (int, optional): Maximum length of the generated sequences. If None, it will be set to
                                         the length of the input plus `num`. Defaults to None.
            temperature (float, optional): Sampling temperature. Defaults to 0.7.
            do_sample (bool, optional): Whether to use sampling; use greedy decoding otherwise. Defaults to True.
            **kwargs: Additional generation parameters.

        Returns:
            Tuple[List[str], List[str]]: A tuple containing the original prompts and their generated continuations.
        """
        # Ensure the tokenizer has a pad_token
        if self.tokenizer.pad_token is not None:
            pass
        elif self.tokenizer.eos_token is not None:
            self.tokenizer.pad_token = self.tokenizer.eos_token
        else:
            raise ValueError("Tokenizer has neither pad_token nor eos_token defined.")

        # Tokenize all prompts in the batch
        batch_encodings = self.tokenizer(
            batch_prompts,
            padding=True,
            truncation=False,
            max_length=1000,
            return_tensors="pt"
        )
        orig_len = batch_encodings.input_ids.shape[1]

        # Determine the new maximum length
        new_length = orig_len + num if max_length is None else max_length

        # Generate outputs
        generate_ids = self.predictor.generate(
            input_ids=batch_encodings.input_ids.to(self.device),
            attention_mask=batch_encodings.attention_mask.to(self.device),
            max_length=new_length,
            do_sample=do_sample,
            temperature=temperature,
            pad_token_id=self.tokenizer.pad_token_id,
            **kwargs,
        )

        # Decode all generated sequences at once
        batch_text_after = self.tokenizer.batch_decode(
            [ids[orig_len:] for ids in generate_ids],
            skip_special_tokens=True,
            clean_up_tokenization_spaces=False
        )

        return batch_prompts, batch_text_after

    # Get intermediate activaitons (using HOOKS!!!)
    def get_midlayer_activations(self, text=None, input_ids=None, raw_img=None, pixel_values=None):
        # Set up the activations to be collected minimally
        self.hooks.disable_all_collect_hooks()
        self.hooks.enable_collect_hooks(["mlp_pre_out", "attn_pre_out"])

        # Run model
        _outputs_embeds = self.get_outputs_embeds(text, input_ids, raw_img, pixel_values)

        # Collect and return activaitons
        return {
            "attn": self.collect_recent_attn_pre_out(),
            "mlp":  self.collect_recent_mlp_pre_out(),
        }

    def collect_recent_mlp_pre_out(self):
        mlp_activations = torch.stack([x.to(self.device) for x in self.hooks["mlp_pre_out"]["collect"]])
        mlp_activations = einops.rearrange(mlp_activations,
            "layer batch token dim -> batch layer token dim")
        return mlp_activations

    def collect_recent_attn_pre_out(self):
        attn_activations = torch.stack([x.to(self.device) for x in self.hooks["attn_pre_out"]["collect"]])
        attn_activations = einops.rearrange(attn_activations,
            "layer batch token n_heads d_head -> batch layer token n_heads d_head",
            n_heads=self.cfg.n_heads, d_head=self.cfg.d_head
        )
        return attn_activations

    def get_residual_stream_decoder(self, text):
        self.hooks.collect_hooks()
        self.hooks.enable_collect_hooks(["post_decoder"])
        # Forward pass
        inputs_embeds = self.get_inputs_embeds(text)
        outputs_embeds = self.get_outputs_embeds(inputs_embeds=inputs_embeds)
        # Collect residual stream
        post_decoder_activations = self.hooks["post_decoder"]["collect"]

        return einops.rearrange(
            torch.stack([inputs_embeds, *post_decoder_activations]),
            "layer batch token dim -> batch layer token dim"
        )


    def get_residual_stream(self, text, split=False, decoder_blocks=False):
        if decoder_blocks:
            return self.get_residual_stream_decoder(text)
        # Set up the activations to be collected minimally
        self.hooks.disable_all_collect_hooks()
        self.hooks.enable_collect_hooks(["pre_mlp", "pre_attn"])

        # Forward pass
        outputs_embeds = self.get_outputs_embeds(text)

        # Collect residual stream
        pre_attn_activations = self.hooks["pre_attn"]["collect"]
        pre_mlp_activations  = self.hooks["pre_mlp"]["collect"]

        residual_stream = []
        for pre_attn, pre_mlp in zip(pre_attn_activations, pre_mlp_activations):
            layer_residuals = []
            if pre_attn is not None:
                layer_residuals.append(pre_attn)
            if pre_mlp is not None:
                layer_residuals.append(pre_mlp)
            if layer_residuals:
                if not split:
                    residual_stream += layer_residuals
                else:
                    residual_stream.append(torch.stack(layer_residuals))

        if residual_stream is None:
            print("WARNING: Could not get residual stream.")
            return None

        if split:
            return einops.rearrange(torch.stack(residual_stream),
                "layer component batch token dim -> component batch layer token dim"
            )

        residual_stream.append(outputs_embeds)
        return einops.rearrange(torch.stack(residual_stream),
            "layer batch token dim -> batch layer token dim"
        )

    def get_residual_diffs(self, text=None, input_ids=None, split=True):
        self.hooks.disable_all_collect_hooks()
        self.hooks.enable_collect_hooks(["post_attn", "post_mlp"])

        # Forward pass
        inputs_embeds  = self.get_inputs_embeds(text, input_ids)
        outputs_embeds = self.get_outputs_embeds(inputs_embeds=inputs_embeds)

        # Collect residual stream
        post_attn_activations = self.hooks["post_attn"]["collect"]
        post_mlp_activations  = self.hooks["post_mlp"]["collect"]
        post_attn_activations = einops.rearrange(
            torch.stack(post_attn_activations),
            "layer batch token dim -> batch layer token dim")
        post_mlp_activations  = einops.rearrange(
            torch.stack(post_mlp_activations),
            "layer batch token dim -> batch layer token dim")

        return inputs_embeds, post_attn_activations, post_mlp_activations, outputs_embeds

    def get_text_activations(self, text=None, input_ids=None):
        return self.get_residual_diffs(text=text, input_ids=input_ids)

    def components_loop(self):
        yield self.predictor
        yield self.model
        for hook in self.hooks.all_hooks:
            yield hook

    def to( self, device ):
        if self.use_accelerator: # If using accelerator, init handles multi-device
            return
        if self.dtype_map.is_low_precision: # 8bit & 4bit mode handled by accelerator
            return
        if self.use_quantization:
            return
        orig_device = self.device
        self.device = device
        if self.output_device == orig_device:
            self.output_device = self.device
        for component in self.components_loop():
            component.to(self.device)
        return self

    def forward(self, input_ids):
        input_ids = input_ids.to(self.device)
        return self.model(input_ids)



import einops
import torch

from transformer_lens.HookedTransformerConfig import HookedTransformerConfig


def convert_neo_weights(neo, cfg: HookedTransformerConfig):
    state_dict = {}

    state_dict["embed.W_E"] = neo.transformer.wte.weight
    state_dict["pos_embed.W_pos"] = neo.transformer.wpe.weight

    for l in range(cfg.n_layers):
        state_dict[f"blocks.{l}.ln1.w"] = neo.transformer.h[l].ln_1.weight
        state_dict[f"blocks.{l}.ln1.b"] = neo.transformer.h[l].ln_1.bias

        W_Q = neo.transformer.h[l].attn.attention.q_proj.weight
        W_K = neo.transformer.h[l].attn.attention.k_proj.weight
        W_V = neo.transformer.h[l].attn.attention.v_proj.weight
        W_Q = einops.rearrange(W_Q, "(i h) m->i m h", i=cfg.n_heads)
        W_K = einops.rearrange(W_K, "(i h) m->i m h", i=cfg.n_heads)
        W_V = einops.rearrange(W_V, "(i h) m->i m h", i=cfg.n_heads)
        state_dict[f"blocks.{l}.attn.W_Q"] = W_Q
        state_dict[f"blocks.{l}.attn.W_K"] = W_K
        state_dict[f"blocks.{l}.attn.W_V"] = W_V

        state_dict[f"blocks.{l}.attn.b_Q"] = torch.zeros(cfg.n_heads, cfg.d_head, dtype=cfg.dtype)
        state_dict[f"blocks.{l}.attn.b_K"] = torch.zeros(cfg.n_heads, cfg.d_head, dtype=cfg.dtype)
        state_dict[f"blocks.{l}.attn.b_V"] = torch.zeros(cfg.n_heads, cfg.d_head, dtype=cfg.dtype)

        W_O = neo.transformer.h[l].attn.attention.out_proj.weight
        W_O = einops.rearrange(W_O, "m (i h)->i h m", i=cfg.n_heads)
        state_dict[f"blocks.{l}.attn.W_O"] = W_O
        state_dict[f"blocks.{l}.attn.b_O"] = neo.transformer.h[l].attn.attention.out_proj.bias

        state_dict[f"blocks.{l}.ln2.w"] = neo.transformer.h[l].ln_2.weight
        state_dict[f"blocks.{l}.ln2.b"] = neo.transformer.h[l].ln_2.bias

        state_dict[f"blocks.{l}.mlp.W_in"] = neo.transformer.h[l].mlp.c_fc.weight.T
        state_dict[f"blocks.{l}.mlp.b_in"] = neo.transformer.h[l].mlp.c_fc.bias

        state_dict[f"blocks.{l}.mlp.W_out"] = neo.transformer.h[l].mlp.c_proj.weight.T
        state_dict[f"blocks.{l}.mlp.b_out"] = neo.transformer.h[l].mlp.c_proj.bias
    state_dict["ln_final.w"] = neo.transformer.ln_f.weight
    state_dict["ln_final.b"] = neo.transformer.ln_f.bias

    state_dict["unembed.W_U"] = neo.lm_head.weight.T
    state_dict["unembed.b_U"] = torch.zeros(cfg.d_vocab, dtype=cfg.dtype)
    return state_dict



"""Evaluation Helpers.

This module contains some rough evals for models, but you are likely better off using the
HuggingFace evaluate library if you want to do anything properly. This is however here if you want
it and want to eg cheaply and roughly compare models you've trained to baselines.
"""

import random
from typing import Dict, List, Optional

import einops
import torch
import tqdm.auto as tqdm
from datasets import load_dataset
from torch.utils.data import DataLoader, Dataset

from transformer_lens import utils


# %%
def sanity_check(model):
    """
    Very basic eval - just feeds a string into the model (in this case, the first paragraph of Circuits: Zoom In), and returns the loss. It's a rough and quick sanity check - if the loss is <5 the model is probably OK, if the loss is >7 something's gone wrong.

    Note that this is a very basic eval, and doesn't really tell you much about the model's performance.
    """

    text = "Many important transition points in the history of science have been moments when science 'zoomed in.' At these points, we develop a visualization or tool that allows us to see the world in a new level of detail, and a new field of science develops to study the world through this lens."

    return model(text, return_type="loss")


# %%
def make_wiki_data_loader(tokenizer, batch_size=8):
    """
    Evaluate on Wikitext 2, a dump of Wikipedia articles. (Using the train set because it's larger, I don't really expect anyone to bother with quarantining the validation set nowadays.)

    Note there's likely to be dataset leakage into training data (though I believe GPT-2 was explicitly trained on non-Wikipedia data)
    """
    wiki_data = load_dataset("wikitext", "wikitext-2-v1", split="train")
    print(len(wiki_data))
    dataset = utils.tokenize_and_concatenate(wiki_data, tokenizer)
    data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True, drop_last=True)
    return data_loader


def make_owt_data_loader(tokenizer, batch_size=8):
    """
    Evaluate on OpenWebText an open source replication of the GPT-2 training corpus (Reddit links with >3 karma)

    I think the Mistral models were trained on this dataset, so they get very good performance.
    """
    owt_data = load_dataset("stas/openwebtext-10k", split="train")
    print(len(owt_data))
    dataset = utils.tokenize_and_concatenate(owt_data, tokenizer)
    data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True, drop_last=True)
    return data_loader


def make_pile_data_loader(tokenizer, batch_size=8):
    """
    Evaluate on the first 10k texts from The Pile.

    The Pile is EleutherAI's general-purpose english dataset, made of 22 subsets
    including academic papers, books, internet content...
    """
    pile_data = load_dataset("NeelNanda/pile-10k", split="train")
    print(len(pile_data))
    dataset = utils.tokenize_and_concatenate(pile_data, tokenizer)
    data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True, drop_last=True)
    return data_loader


def make_code_data_loader(tokenizer, batch_size=8):
    """
    Evaluate on the CodeParrot dataset, a dump of Python code.

    All models seem to get significantly lower loss here (even non-code trained models like GPT-2),
    presumably code is much easier to predict than natural language?
    """
    code_data = load_dataset("codeparrot/codeparrot-valid-v2-near-dedup", split="train")
    print(len(code_data))
    dataset = utils.tokenize_and_concatenate(code_data, tokenizer, column_name="content")
    data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True, drop_last=True)
    return data_loader


DATASET_NAMES = ["wiki", "owt", "pile", "code"]
DATASET_LOADERS = [
    make_wiki_data_loader,
    make_owt_data_loader,
    make_pile_data_loader,
    make_code_data_loader,
]


# %%
@torch.inference_mode()
def evaluate_on_dataset(model, data_loader, truncate=100, device="cuda"):
    running_loss = 0
    total = 0
    for batch in tqdm.tqdm(data_loader):
        loss = model(batch["tokens"].to(device), return_type="loss").mean()
        running_loss += loss.item()
        total += 1
        if total > truncate:
            break
    return running_loss / total


# %%
@torch.inference_mode()
def induction_loss(
    model, tokenizer=None, batch_size=4, subseq_len=384, prepend_bos=None, device="cuda"
):
    """
    Generates a batch of random sequences repeated twice, and measures model performance on the second half. Tests whether a model has induction heads.

    By default, prepends a beginning of string token (when prepend_bos flag defaults to None, model.cfg.default_prepend_bos is used
    whose default is True unless specified otherwise), which is useful to give models a resting position, and sometimes models were trained with this.
    """
    # Make the repeated sequence
    first_half_tokens = torch.randint(100, 20000, (batch_size, subseq_len)).to(device)
    repeated_tokens = einops.repeat(first_half_tokens, "b p -> b (2 p)")

    # Use the provided prepend_bos as an override if it's not None;
    # otherwise use model.cfg.default_prepend_bos (defaults to True)
    prepend_bos = utils.override_or_use_default_value(
        model.cfg.default_prepend_bos, override=prepend_bos
    )

    # Prepend a Beginning Of String token
    if prepend_bos:
        if tokenizer is None:
            tokenizer = model.tokenizer
        repeated_tokens[:, 0] = tokenizer.bos_token_id
    # Run the model, and extract the per token correct log prob
    logits = model(repeated_tokens, return_type="logits")
    correct_log_probs = utils.lm_cross_entropy_loss(logits, repeated_tokens, per_token=True)
    # Take the loss over the second half of the sequence
    return correct_log_probs[:, subseq_len + 1 :].mean()


# %%
@torch.inference_mode()
def evaluate(model, truncate=100, batch_size=8, tokenizer=None):
    if tokenizer is None:
        tokenizer = model.tokenizer
    losses = {}
    for data_name, data_loader_fn in zip(DATASET_NAMES, DATASET_LOADERS):
        data_loader = data_loader_fn(tokenizer=tokenizer, batch_size=batch_size)
        loss = evaluate_on_dataset(model, data_loader, truncate=truncate)
        print(f"{data_name}: {loss}")
        losses[f"{data_name}_loss"] = loss
    return losses


# %%
class IOIDataset(Dataset):
    """
    Dataset for Indirect Object Identification tasks.
    Paper: https://arxiv.org/pdf/2211.00593.pdf

    Example:

    .. code-block:: python

        >>> from transformer_lens.evals import ioi_eval, IOIDataset
        >>> from transformer_lens.HookedTransformer import HookedTransformer

        >>> model = HookedTransformer.from_pretrained('gpt2-small')
        Loaded pretrained model gpt2-small into HookedTransformer

        >>> # Evaluate like this, printing the logit difference
        >>> print(round(ioi_eval(model, num_samples=100)["Logit Difference"], 3))
        5.476

        >>> # Can use custom dataset
        >>> ds = IOIDataset(
        ...     tokenizer=model.tokenizer,
        ...     num_samples=100,
        ...     templates=['[A] met with [B]. [B] gave the [OBJECT] to [A]'],
        ...     names=['Alice', 'Bob', 'Charlie'],
        ...     nouns={'OBJECT': ['ball', 'book']},
        ... )
        >>> print(round(ioi_eval(model, dataset=ds)["Logit Difference"], 3))
        5.397
    """

    def __init__(
        self,
        tokenizer,
        templates: Optional[List[str]] = None,
        names: Optional[List[str]] = None,
        nouns: Optional[Dict[str, List[str]]] = None,
        num_samples: int = 1000,
        symmetric: bool = False,
        prepend_bos: bool = True,
    ):
        self.tokenizer = tokenizer
        self.prepend_bos = prepend_bos

        self.templates = templates if templates is not None else self.get_default_templates()
        self.names = names if names is not None else self.get_default_names()
        self.nouns = nouns if nouns is not None else self.get_default_nouns()

        self.samples = []
        for _ in range(num_samples // 2 if symmetric else num_samples):
            # If symmetric, get_sample will return two samples
            self.samples.extend(self.get_sample(symmetric=symmetric))

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        sample = self.samples[idx]
        prompt = self.tokenizer.encode(sample["text"])
        if self.prepend_bos:
            prompt = [self.tokenizer.bos_token_id] + prompt

        return {
            "prompt": torch.LongTensor(prompt),
            "IO": torch.LongTensor(self.tokenizer.encode(sample["IO"])),
            "S": torch.LongTensor(self.tokenizer.encode(sample["S"])),
        }

    def get_sample(self, symmetric=False) -> List[Dict[str, str]]:
        random.seed(42)
        template: str = random.choice(self.templates)
        for noun_type, noun_list in self.nouns.items():
            template = template.replace(f"[{noun_type}]", random.choice(noun_list))

        samples: List[Dict[str, str]] = []

        # Sample two names without replacement
        names = random.sample(self.names, 2)
        sample = template.replace("[A]", names[0])
        sample = sample.replace("[B]", names[1])
        # Prepend spaces to IO and S so that the target is e.g. " Mary" and not "Mary"
        samples.append({"text": sample, "IO": " " + names[0], "S": " " + names[1]})

        if symmetric:
            sample_2 = template.replace("[A]", names[1])
            sample_2 = sample_2.replace("[B]", names[0])
            samples.append({"text": sample_2, "IO": " " + names[1], "S": " " + names[0]})

        return samples

    @staticmethod
    def get_default_names():
        return ["John", "Mary"]

    @staticmethod
    def get_default_templates():
        return [
            "[A] and [B] went to the [LOCATION] to buy [OBJECT]. [B] handed the [OBJECT] to [A]",
            "Then, [B] and [A] went to the [LOCATION]. [B] gave the [OBJECT] to [A]",
        ]

    @staticmethod
    def get_default_nouns():
        return {
            "LOCATION": ["store", "market"],
            "OBJECT": ["milk", "eggs", "bread"],
        }


@torch.inference_mode()
def ioi_eval(model, dataset=None, batch_size=8, num_samples=1000, tokenizer=None, symmetric=False):
    """Evaluate the Model on the Indirect Object Identification Task.

    Args:
        model: HookedTransformer model.
        dataset: PyTorch Dataset that returns a dict with keys "prompt", "IO", and "S".
        batch_size: Batch size to use.
        num_samples: Number of samples to use.
        tokenizer: Tokenizer to use.
        symmetric: Whether to use the symmetric version of the task.

    Returns:
        Average logit difference and accuracy.
    """
    if tokenizer is None:
        tokenizer = model.tokenizer

    if dataset is None:
        dataset = IOIDataset(tokenizer, num_samples=num_samples, symmetric=symmetric)

    def collate(samples):
        prompts = [sample["prompt"] for sample in samples]
        padded_prompts = torch.nn.utils.rnn.pad_sequence(prompts, batch_first=True)
        return {
            "prompt": padded_prompts,
            "IO": [sample["IO"] for sample in samples],
            "S": [sample["S"] for sample in samples],
            "prompt_length": [p.shape[0] for p in prompts],
        }

    data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True, collate_fn=collate)

    total_correct = 0
    total_logit_diff = 0
    for batch in tqdm.tqdm(data_loader):
        batch_logits = model(batch["prompt"], return_type="logits")

        for i in range(batch_logits.shape[0]):
            io = batch["IO"][i]
            s = batch["S"][i]
            prefix_length = batch["prompt_length"][i] - io.shape[0]

            # Trim io and s to the same length
            min_len = min(io.shape[0], s.shape[0])
            io = io[:min_len]
            s = s[:min_len]

            # Remove identical prefixes
            start_idx = torch.where(io != s)[0][0]
            io = io[start_idx]
            s = s[start_idx]
            logit_idx = prefix_length + start_idx - 1

            # Get the logits for the tokens we care about
            logits = batch_logits[i, logit_idx]
            correct_logit = logits[io]
            incorrect_logit = logits[s]

            # Compute stats
            logit_diff = correct_logit - incorrect_logit
            correct = logit_diff > 0
            total_correct += correct.item()
            total_logit_diff += logit_diff.item()

    return {
        "Logit Difference": total_logit_diff / len(dataset),
        "Accuracy": total_correct / len(dataset),
    }



import math
from typing import Dict, Optional, Union

import torch
import torch.nn as nn
from jaxtyping import Float, Int

from transformer_lens.components.abstract_attention import AbstractAttention
from transformer_lens.hook_points import HookPoint
from transformer_lens.HookedTransformerConfig import HookedTransformerConfig


class T5Attention(AbstractAttention):
    r"""
    T5 attention - with relative attention bias and cross-attention support
    This realisation expects you to precompute relative positional bias, and then feed it to forward
    like
    ```python
    attn = T5Attention(cfg, has_relative_attention_bias=True)
    positional_bias = attn.compute_relative_attention_bias(query_len, key_len, device=device)
    result = attn(query, key, value, position_bias=positional_bias)
    ```
    """

    def __init__(
        self,
        cfg: Union[Dict, HookedTransformerConfig],
        has_relative_attention_bias: bool = False,
        attn_type: str = "global",
        layer_id: Optional[int] = None,
    ):
        super().__init__(cfg, attn_type, layer_id)
        if isinstance(cfg, Dict):
            cfg = HookedTransformerConfig.from_dict(cfg)
        self.cfg = cfg
        self.has_relative_attention_bias: bool = has_relative_attention_bias

        if self.has_relative_attention_bias:
            if (
                cfg.relative_attention_num_buckets is None
                or cfg.relative_attention_max_distance is None
            ):
                raise ValueError(
                    "You need to specify relative_attention_num_buckets and relative_attention_max_distance  in config to use relative attention bias"
                )

            self.relative_attention_num_buckets = cfg.relative_attention_num_buckets
            self.relative_attention_max_distance = cfg.relative_attention_max_distance
            self.rel_pos_bias = nn.Embedding(self.relative_attention_num_buckets, self.cfg.n_heads)
            self.rel_pos_hook = HookPoint()

        self.W_K = nn.Parameter(
            torch.empty(self.cfg.n_heads, self.cfg.d_model, self.cfg.d_head, dtype=cfg.dtype)
        )
        self.W_V = nn.Parameter(
            torch.empty(self.cfg.n_heads, self.cfg.d_model, self.cfg.d_head, dtype=cfg.dtype)
        )
        self.b_K = nn.Parameter(torch.zeros(self.cfg.n_heads, self.cfg.d_head, dtype=cfg.dtype))
        self.b_V = nn.Parameter(torch.zeros(self.cfg.n_heads, self.cfg.d_head, dtype=cfg.dtype))

    @staticmethod
    def _relative_position_bucket(
        relative_position: Int[torch.Tensor, "query_pos kv_pos"],
        bidirectional=True,
        num_buckets=32,
        max_distance=128,
    ) -> Int[torch.Tensor, "query_pos kv_pos"]:
        """
        added from
        https://github.com/huggingface/transformers/blob/e0c3cee17085914bbe505c159beeb8ae39bc37dd/src/transformers/models/t5/modeling_t5.py#L382
        which is adapted from
        https://github.com/tensorflow/mesh/blob/0cb87fe07da627bf0b7e60475d59f95ed6b5be3d/mesh_tensorflow/transformer/transformer_layers.py#L593


        Translate relative position to a bucket number for relative attention. The relative position is defined as
        memory_position - query_position, i.e. the distance in tokens from the attending position to the attended-to
        position. If bidirectional=False, then positive relative positions are invalid. We use smaller buckets for
        small absolute relative_position and larger buckets for larger absolute relative_positions. All relative
        positions >=max_distance map to the same bucket. All relative positions <=-max_distance map to the same bucket.
        This should allow for more graceful generalization to longer sequences than the model has been trained on

        Args:
            relative_position: an int32 Tensor
            bidirectional: a boolean - whether the attention is bidirectional
            num_buckets: an integer
            max_distance: an integer

        Returns:
            a Tensor with the same shape as relative_position, containing int32 values in the range [0, num_buckets)
        """
        relative_buckets = torch.zeros_like(relative_position)

        if bidirectional:
            num_buckets //= 2
            relative_buckets += (relative_position > 0).to(torch.long) * num_buckets
            relative_position = torch.abs(relative_position)
        else:
            relative_position = -torch.min(relative_position, torch.zeros_like(relative_position))
        # now relative_position is in the range [0, inf)

        # half of the buckets are for exact increments in positions
        max_exact = num_buckets // 2
        is_small = relative_position < max_exact

        # The other half of the buckets are for logarithmically bigger bins in positions up to max_distance
        relative_position_if_large = max_exact + (
            torch.log(relative_position.float() / max_exact)
            / math.log(max_distance / max_exact)
            * (num_buckets - max_exact)
        ).to(torch.long)
        relative_position_if_large = torch.min(
            relative_position_if_large,
            torch.full_like(relative_position_if_large, num_buckets - 1),
        )

        relative_buckets += torch.where(is_small, relative_position, relative_position_if_large)
        return relative_buckets

    def compute_relative_attention_bias(
        self, query_length: int, key_length: int, device=None
    ) -> Float[torch.Tensor, "1 head_index pos kv_pos"]:
        """Compute binned relative position bias"""
        if device is None:
            device = self.rel_pos_bias.weight.device
        context_position = torch.arange(query_length, dtype=torch.long, device=device)[:, None]
        memory_position = torch.arange(key_length, dtype=torch.long, device=device)[None, :]
        relative_position = memory_position - context_position  # shape (query_length, key_length)
        relative_position_bucket = self._relative_position_bucket(
            relative_position,  # shape (query_length, key_length)
            bidirectional=True,
            num_buckets=self.relative_attention_num_buckets,
            max_distance=self.relative_attention_max_distance,
        )
        values = self.rel_pos_bias(
            relative_position_bucket
        )  # shape (query_length, key_length, num_heads)
        values = values.permute([2, 0, 1]).unsqueeze(
            0
        )  # shape (1, num_heads, query_length, key_length)
        return values



"""Next Sentence Prediction.

Contains a BERT style model specifically for Next Sentence Prediction. This is separate from
:class:`transformer_lens.HookedTransformer` because it has a significantly different architecture
to e.g. GPT style transformers.
"""

from typing import Dict, List, Optional, Tuple, Union, overload

import torch
from jaxtyping import Float, Int
from typing_extensions import Literal

from transformer_lens.ActivationCache import ActivationCache
from transformer_lens.HookedEncoder import HookedEncoder


class BertNextSentencePrediction:
    """A BERT-style model for Next Sentence Prediction (NSP) that extends HookedEncoder.

    This class implements a BERT model specifically designed for the Next Sentence Prediction task,
    where the model predicts whether two input sentences naturally follow each other in the original text.
    It inherits from HookedEncoder and adds NSP-specific components like the NSP head and pooler layer.

    The model processes pairs of sentences and outputs either logits or human-readable predictions
    indicating whether the sentences are sequential. String inputs are automatically tokenized with
    appropriate token type IDs to distinguish between the two sentences.

    Note:
        This model expects inputs to be provided as pairs of sentences. Single sentence inputs
        or inputs without proper sentence separation will raise errors.
    """

    def __init__(self, model: HookedEncoder):
        self.model = model

    def __call__(
        self,
        input: Union[
            List[str],
            Int[torch.Tensor, "batch pos"],
        ],
        return_type: Optional[Union[Literal["logits"], Literal["predictions"]]] = "logits",
        token_type_ids: Optional[Int[torch.Tensor, "batch pos"]] = None,
        one_zero_attention_mask: Optional[Int[torch.Tensor, "batch pos"]] = None,
    ) -> Optional[Union[Float[torch.Tensor, "batch 2"], str]]:
        """Makes the NextSentencePrediction instance callable.

        This method delegates to the forward method, allowing the model to be called directly.
        The arguments and return types match the forward method exactly.
        """
        return self.forward(
            input,
            return_type=return_type,
            token_type_ids=token_type_ids,
            one_zero_attention_mask=one_zero_attention_mask,
        )

    def to_tokens(
        self,
        input: List[str],
        move_to_device: bool = True,
        truncate: bool = True,
    ) -> Tuple[
        Int[torch.Tensor, "batch pos"],
        Int[torch.Tensor, "batch pos"],
        Int[torch.Tensor, "batch pos"],
    ]:
        """Converts a string to a tensor of tokens.
        Taken mostly from the HookedTransformer implementation, but does not support default padding
        sides or prepend_bos.
        Args:
            input: List[str]]: The input to tokenize.
            move_to_device (bool): Whether to move the output tensor of tokens to the device the model lives on. Defaults to True
            truncate (bool): If the output tokens are too long, whether to truncate the output
            tokens to the model's max context window. Does nothing for shorter inputs. Defaults to
            True.
        """

        if len(input) != 2:
            raise ValueError(
                "Next sentence prediction task requires exactly two sentences, please provide a list of strings with each sentence as an element."
            )

        # We need to input the two sentences separately for NSP
        encodings = self.model.tokenizer(
            input[0],
            input[1],
            return_tensors="pt",
            padding=True,
            truncation=truncate,
            max_length=self.model.cfg.n_ctx if truncate else None,
        )

        tokens = encodings["input_ids"]
        token_type_ids = encodings["token_type_ids"]
        attention_mask = encodings["attention_mask"]

        if move_to_device:
            tokens = tokens.to(self.model.cfg.device)
            token_type_ids = token_type_ids.to(self.model.cfg.device)
            attention_mask = attention_mask.to(self.model.cfg.device)

        return tokens, token_type_ids, attention_mask

    @overload
    def forward(
        self,
        input: Union[
            List[str],
            Int[torch.Tensor, "batch pos"],
        ],
        return_type: Union[Literal["logits"], Literal["predictions"]],
        token_type_ids: Optional[Int[torch.Tensor, "batch pos"]] = None,
        one_zero_attention_mask: Optional[Int[torch.Tensor, "batch pos"]] = None,
    ) -> Union[Float[torch.Tensor, "batch 2"], str]:
        ...

    @overload
    def forward(
        self,
        input: Union[
            List[str],
            Int[torch.Tensor, "batch pos"],
        ],
        return_type: Literal[None],
        token_type_ids: Optional[Int[torch.Tensor, "batch pos"]] = None,
        one_zero_attention_mask: Optional[Int[torch.Tensor, "batch pos"]] = None,
    ) -> Optional[Union[Float[torch.Tensor, "batch 2"], str]]:
        ...

    def forward(
        self,
        input: Union[
            List[str],
            Int[torch.Tensor, "batch pos"],
        ],
        return_type: Optional[Union[Literal["logits"], Literal["predictions"]]] = "logits",
        token_type_ids: Optional[Int[torch.Tensor, "batch pos"]] = None,
        one_zero_attention_mask: Optional[Int[torch.Tensor, "batch pos"]] = None,
    ) -> Optional[Union[Float[torch.Tensor, "batch 2"], str]]:
        """Forward pass through the NextSentencePrediction module. Performs Next Sentence Prediction on a pair of sentences.

        Args:
            input: The input to process. Can be one of:
                - List[str]: A list of two strings representing the two sentences NSP should be performed on
                - torch.Tensor: Input tokens as integers with shape (batch, position)
            return_type: Optional[str]: The type of output to return. Can be one of:
                - None: Return nothing, don't calculate logits
                - 'logits': Return logits tensor
                - 'predictions': Return human-readable predictions
            token_type_ids: Optional[torch.Tensor]: Binary ids indicating whether a token belongs
                to sequence A or B. For example, for two sentences:
                "[CLS] Sentence A [SEP] Sentence B [SEP]", token_type_ids would be
                [0, 0, ..., 0, 1, ..., 1, 1]. `0` represents tokens from Sentence A,
                `1` from Sentence B. If not provided, BERT assumes a single sequence input.
                This parameter gets inferred from the the tokenizer if input is a string or list of strings.
                Shape is (batch_size, sequence_length).
            one_zero_attention_mask: Optional[torch.Tensor]: A binary mask which indicates
                which tokens should be attended to (1) and which should be ignored (0).
                Primarily used for padding variable-length sentences in a batch.
                For instance, in a batch with sentences of differing lengths, shorter
                sentences are padded with 0s on the right. If not provided, the model
                assumes all tokens should be attended to.
                This parameter gets inferred from the tokenizer if input is a string or list of strings.
                Shape is (batch_size, sequence_length).

        Returns:
            Optional[torch.Tensor]: Depending on return_type:
                - None: Returns None if return_type is None
                - torch.Tensor: Returns logits if return_type is 'logits' (or if return_type is not explicitly provided)
                    - Shape is (batch_size, 2)
                - str or List[str]: Returns string indicating if sentences are sequential if return_type is 'predictions'

        Raises:
            ValueError: If using NSP task without proper input format or token_type_ids
            AssertionError: If using string input without a tokenizer
        """

        if isinstance(input, list):
            assert self.model.tokenizer is not None, "Must provide a tokenizer if input is a string"
            tokens, token_type_ids_from_tokenizer, attention_mask = self.to_tokens(input)

            # If token_type_ids or attention mask are not provided, use the ones from the tokenizer
            token_type_ids = (
                token_type_ids_from_tokenizer if token_type_ids is None else token_type_ids
            )
            one_zero_attention_mask = (
                attention_mask if one_zero_attention_mask is None else one_zero_attention_mask
            )
        elif token_type_ids == None and isinstance(input, torch.Tensor):
            raise ValueError(
                "You are using the NSP task without specifying token_type_ids."
                "This means that the model will treat the input as a single sequence which will lead to incorrect results."
                "Please provide token_type_ids or use a string input."
            )
        else:
            tokens = input

        resid = self.model.encoder_output(tokens, token_type_ids, one_zero_attention_mask)

        # NSP requires pooling (for more information see BertPooler)
        resid = self.model.pooler(resid)
        logits = self.model.nsp_head(resid)

        if return_type == "predictions":
            logprobs = logits.log_softmax(dim=-1)
            predictions = [
                "The sentences are sequential",
                "The sentences are NOT sequential",
            ]
            return predictions[logprobs.argmax(dim=-1).item()]

        elif return_type == None:
            return None

        return logits

    @overload
    def run_with_cache(
        self, *model_args, return_cache_object: Literal[True] = True, **kwargs
    ) -> Tuple[Float[torch.Tensor, "batch 2"], ActivationCache,]:
        ...

    @overload
    def run_with_cache(
        self, *model_args, return_cache_object: Literal[False], **kwargs
    ) -> Tuple[Float[torch.Tensor, "batch 2"], Dict[str, torch.Tensor],]:
        ...

    def run_with_cache(
        self,
        *model_args,
        return_cache_object: bool = True,
        remove_batch_dim: bool = False,
        **kwargs,
    ) -> Tuple[Float[torch.Tensor, "batch 2"], Union[ActivationCache, Dict[str, torch.Tensor]],]:
        """
        Wrapper around run_with_cache in HookedRootModule. If return_cache_object is True,
        this will return an ActivationCache object, with a bunch of useful HookedTransformer specific methods,
        otherwise it will return a dictionary of activations as in HookedRootModule.
        This function was copied directly from HookedTransformer.
        """

        # Create wrapper for forward function, such that run_with_cache uses
        # the forward function of this class and not HookedEncoder

        class ForwardWrapper:
            def __init__(self, nsp):
                self.nsp = nsp
                self.original_forward = nsp.model.forward

            def __enter__(self):
                # Store reference to wrapper function
                def wrapped_forward(*fargs, **fkwargs):
                    return self.nsp.forward(*fargs, **fkwargs)

                self.nsp.model.forward = wrapped_forward
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                # Restore original forward
                self.nsp.model.forward = self.original_forward

        with ForwardWrapper(self):
            out, cache_dict = self.model.run_with_cache(
                *model_args, remove_batch_dim=remove_batch_dim, **kwargs
            )
            if return_cache_object:
                cache = ActivationCache(cache_dict, self, has_batch_dim=not remove_batch_dim)
                return out, cache
            else:
                return out, cache_dict



"""Past Key Value Caching.

This module contains the HookedTransformerKeyValueCache and HookedTransformerKeyValueCacheEntry
classes, which are used to store past keys and values for the Transformer. This is important for
generating text - we can cache a lot of past computation and avoid repeating ourselves!
"""
from dataclasses import dataclass
from typing import List, Union

import torch
from jaxtyping import Float, Int

from transformer_lens.HookedTransformerConfig import HookedTransformerConfig
from transformer_lens.utilities.devices import get_device_for_block_index


@dataclass
class HookedTransformerKeyValueCacheEntry:
    past_keys: Float[torch.Tensor, "batch pos_so_far n_heads d_head"]
    past_values: Float[torch.Tensor, "batch pos_so_far n_heads d_head"]
    frozen: bool = False

    @classmethod
    def init_cache_entry(
        cls,
        cfg: HookedTransformerConfig,
        device: Union[torch.device, str, None],
        batch_size: int = 1,
    ):
        n_heads = cfg.n_key_value_heads if cfg.n_key_value_heads is not None else cfg.n_heads
        return cls(
            past_keys=torch.empty(
                (batch_size, 0, n_heads, cfg.d_head), device=device, dtype=cfg.dtype
            ),
            past_values=torch.empty(
                (batch_size, 0, n_heads, cfg.d_head), device=device, dtype=cfg.dtype
            ),
        )

    def append(
        self,
        new_keys: Float[torch.Tensor, "batch new_tokens n_heads d_head"],
        new_values: Float[torch.Tensor, "batch new_tokens n_heads d_head"],
    ):
        updated_keys: Float[
            torch.Tensor, "batch pos_so_far_plus_new_tokens n_heads d_head"
        ] = torch.cat([self.past_keys, new_keys], dim=1)
        updated_values: Float[
            torch.Tensor, "batch pos_so_far_plus_new_tokens n_heads d_head"
        ] = torch.cat([self.past_values, new_values], dim=1)
        if not self.frozen:
            self.past_keys = updated_keys
            self.past_values = updated_values
        return updated_keys, updated_values


@dataclass
class HookedTransformerKeyValueCache:
    """
    A cache for storing past keys and values for the Transformer. This is important for generating text - we can cache a lot of past computation and avoid repeating ourselves!

    This cache is a list of HookedTransformerKeyValueCacheEntry objects, one for each layer in the Transformer. Each object stores a [batch, pos_so_far, n_heads, d_head] tensor for both keys and values, and each entry has an append method to add a single new key and value.

    The cache can be frozen so that it is not updated during the forward pass. This is useful when we want to run many inputs with the same prefix.
    """

    entries: List[HookedTransformerKeyValueCacheEntry]
    previous_attention_mask: Int[torch.Tensor, "batch pos_so_far"]
    frozen: bool = False

    @classmethod
    def init_cache(
        cls,
        cfg: HookedTransformerConfig,
        device: Union[torch.device, str, None],
        batch_size: int = 1,
    ):
        return cls(
            entries=[
                HookedTransformerKeyValueCacheEntry.init_cache_entry(
                    cfg,
                    get_device_for_block_index(i, cfg, device),
                    batch_size,
                )
                for i in range(cfg.n_layers)
            ],
            previous_attention_mask=torch.empty(
                # This may actually be an int64, but type promotion will handle it:
                # See: https://pytorch.org/docs/stable/tensor_attributes.html#type-promotion-doc
                # See: https://github.com/pytorch/pytorch/issues/35014
                (batch_size, 0),
                device=device,
                dtype=torch.int,
            ),
        )

    def freeze(self):
        self.frozen = True
        for entry in self.entries:
            entry.frozen = True

    def unfreeze(self):
        self.frozen = False
        for entry in self.entries:
            entry.frozen = False

    def append_attention_mask(self, attention_mask: Int[torch.Tensor, "batch new_tokens"]):
        attention_mask = attention_mask.to(self.previous_attention_mask.device)
        updated_attention_mask = torch.cat([self.previous_attention_mask, attention_mask], dim=-1)
        if not self.frozen:
            self.previous_attention_mask = updated_attention_mask
        return updated_attention_mask

    def __getitem__(self, idx):
        return self.entries[idx]



""" This file contains utils for working with single layers of dense neural networks.
"""

from typing import Any, Mapping, Optional, Union, Dict
from torch import Tensor
import torch.nn.functional as F
import torch
from torch.nn.modules.module import Module

######################################################################################
# Define InverseLinear Layer Class
######################################################################################

class InverseLinear(torch.nn.Module):
    """_summary_
    Produces a torch layer which undoes what the "original" linear layer does.

    This was built to get the output right before out_proj, and I thought
    it would be easier and/or faster than using the values matrix
    (it probably is not actually easier or faster)

    Args:
        original: The original Linear layer we are getting the inverse of
    """
    def __init__(self,
            original: torch.nn.Linear=None,
            original_weights: Tensor=None,
            original_biases: Tensor=None,
            n_heads: int = 1,
        ):
        super(InverseLinear, self).__init__()
        if original is not None:
            weights, bias = original.parameters()
        elif original_weights is not None:
            weights = original_weights
            bias    = original_biases
        else:
            raise ValueError("Either original or original_weights must be provided")

        # Check that the original transformation is square
        original_size = weights.size()
        self.n_heads = n_heads
        if original_size[0] != original_size[1]:
            raise ValueError("Original Linear Layer must be square")

        # Define the Inverse Bias vector
        if bias is None:
            bias = torch.zeros(weights.shape[0], device=weights.device)
        self.inverse_bias = - bias.clone()

        # Define the Inverse Linear Layer
        _dtype, _device = weights.dtype, weights.device
        inverse_weights = weights.clone().cpu().to(dtype=torch.float64).inverse()
        inverse_weights = inverse_weights.to(dtype=_dtype).to(_device)
        size = inverse_weights.size()
        self.fc = torch.nn.Linear( size[0], size[1], bias=False )
        self.fc.load_state_dict({'weight': inverse_weights})

    def forward(self, x):
        y = x + self.inverse_bias
        y = self.fc( y )
        return y.reshape([ *x.shape[:-1], self.n_heads, -1 ])

    # pylint: disable=arguments-differ
    def to(self,
           device: Optional[Union[str, torch.device]] = None,
           dtype: Optional[torch.dtype] = None,
           **kwargs
        ):
        super( InverseLinear, self ).to( device, dtype=dtype, **kwargs )
        if device is not None:
            self.inverse_bias = self.inverse_bias.to( device, **kwargs )
            self.fc = self.fc.to( device, **kwargs )
        if dtype is not None:
            self.inverse_bias = self.inverse_bias.to( dtype=dtype, **kwargs )
            self.fc = self.fc.to( dtype=dtype, **kwargs )
        return self

    def state_dict(self):
        return {
            "weight": self.fc.weight,
            "inverse_bias": self.inverse_bias
        }

    def load_state_dict(self, state_dict):
        # Load weight
        fc_params = self.fc.state_dict()
        fc_params["weight"] = state_dict["weight"]
        self.fc.load_state_dict(fc_params)

        # Load inverse bias
        self.inverse_bias = state_dict["inverse_bias"]

######################################################################################
# Define MLP Deletion functions
######################################################################################

def mlp_delete_rows_raw(
        deletion_indices: Tensor,
        weights: Tensor,
        biases: Optional[Tensor] = None,
    ):
    # Delete the weights and biases from the rows
    n_rows = len(weights)
    for row_index in range(n_rows):
        if deletion_indices[row_index]:
            weights[row_index] = torch.zeros_like(weights[row_index])
            if biases is not None:
                biases[row_index]  = torch.zeros_like(biases[row_index])
    return weights, biases


def mlp_delete_rows(mlp:
        torch.nn.Linear,
        deletion_indices: Tensor,
        delete_biases:bool = True,
    ):
    """Deletes (in place) the weights and biases of rows of the MLP that
    are marked True in deletion_rows

    Args:
        mlp (torch.nn.Linear): The Multi-Layer Perceptron to delete rows from
        deletion_rows (Tensor): Tensor of booleans indicating which rows to delete
    """
    # Get the parameters from the MLP
    params = mlp.state_dict()
    weights: Tensor = params['weight']
    biases: Tensor  = params['bias']

    # Delete the weights and biases from the rows
    if delete_biases:
        weights, biases = mlp_delete_rows_raw(deletion_indices, weights, biases)
    else:
        weights, _ = mlp_delete_rows_raw(deletion_indices, weights, None)


    # Update the model to have the deleted rows
    params.update({'weight': weights, 'bias': biases})
    mlp.load_state_dict(params)

    return mlp

def mlp_adjust_biases(
        mlp: torch.nn.Linear,
        deletion_indices: Tensor,
        mean_activations: Tensor,
    ):
    """ Calculates the bias adjustement needed to compensate for the deletion of
    neurons in the MLP, and applies it to the MLP

    Args:
        mlp (torch.nn.Linear): The Multi-Layer Perceptron to adjust the biases of
        deletion_rows (Tensor): The
        mean_activations (Tensor):
    """
    if mean_activations is None:
        return mlp

    # Load parameters
    params = mlp.state_dict()
    biases: Tensor = params['bias']
    device = biases.device

    # Make a temporary copy of the MLP with only the weights (no biases).
    params.update({'bias': torch.zeros_like(biases)})
    mlp.load_state_dict(params)

    # adjust bias of out_proj by mean activations
    mean_activations  = mean_activations.detach().clone().to(device)
    mean_activations *= deletion_indices.to(device).reshape(mean_activations.shape)
    bias_adjustement = mlp( mean_activations )
    biases += bias_adjustement

    # Place biases back into the MLP, with adjustements from above
    params.update({'bias': biases})
    mlp.load_state_dict(params)

    return mlp

def mlp_delete_columns_raw(weights: Tensor, deletion_indices: Tensor):
    """Deletes (in place) the columns of weights in the MLP that are
    marked as True in deletion_indices

    Args:
        weights (Tensor): The MLP Weights to delete columns from
        deletion_indices (Tensor): The indices of the columns to delete
    """
    # Load parameters
    device = weights.device

    # Transpose the weights, then delete row by row
    weights_t = weights.transpose(0, 1)

    for row_index, delete_row in enumerate(deletion_indices):
        if delete_row:
            weights_t[row_index] = torch.zeros_like(weights_t[row_index])

    weights = weights_t.transpose(0, 1)

    return weights

def mlp_delete_columns(mlp: torch.nn.Linear, deletion_indices: Tensor):
    """Deletes (in place) the columns of weights in the MLP that are
    marked as True in deletion_indices

    Args:
        mlp (torch.nn.Linear): The Multi-Layer Perceptron to delete columns from
        deletion_indices (Tensor): The indices of the columns to delete
    """
    # Load parameters
    params = mlp.state_dict()

    # Delete the columns
    weights = mlp_delete_columns_raw(params['weight'], deletion_indices)

    # Update the model to have the deleted columns
    params.update({'weight': weights})
    mlp.load_state_dict(params)

    return mlp

######################################################################################
# Define MLP SVD functions
######################################################################################

def svd_two_layer_raw_heads(
        W_in_heads: Tensor,  # [n_heads, d_head, d_model]
        W_out_heads: Tensor, # [d_model, n_heads, d_head]
        svd_dtype: torch.dtype
    ):
    """
    Performs (in-place) SVD on attn matrices already separated into heads.

    Args:
        W_in_heads (Tensor): [n_heads, d_head, d_model]
        W_out_heads (Tensor): [d_model, n_heads, d_head]
        svd_dtype (torch.dtype): dtype to use for SVD

    Output:
        W_in_heads (Tensor): [n_heads, d_head, d_model]
        W_out_heads (Tensor): [d_model, n_heads, d_head]
    """
    # Get meta information
    dtype = W_in_heads.dtype
    (n_heads, d_head, _d_model) = W_in_heads.shape

    # perform SVD using only the weights
    for head in range(n_heads):
        # Get the weights
        in_weights  = W_in_heads[head]
        out_weights = W_out_heads[:, head]

        # Perform SVD
        big_matrix = torch.matmul(out_weights, in_weights).to(dtype=svd_dtype)
        u_out, s, v_in = torch.linalg.svd(big_matrix, full_matrices=True)

        # Remove rows/columns which we know should have zero rank
        s = s[:d_head].to(dtype=dtype)
        v_in  = v_in[:d_head, :].to(dtype=dtype)  # in_matrix,  eg [64, 768]
        u_out = u_out[:, :d_head].to(dtype=dtype) # out_matrix, eg [768, 64]

        # Scale the v_in and u_out matrices by sqrt(S) each
        s = s.sqrt()
        u_out *= s
        v_in  *= s.unsqueeze(dim=-1)

        # return new weights to original head matrix
        in_weights  = v_in
        out_weights = u_out

        # move to actual original matrices for load_state_dict later
        W_in_heads[head] = in_weights
        W_out_heads[:, head] = out_weights

    return W_in_heads, W_out_heads

def mlp_svd_two_layer_raw(
        W_in: Tensor,
        W_out: Tensor,
        b_in: Tensor,
        b_out: Tensor,
        n_heads: Optional[int] = None,
        svd_dtype: torch.dtype = torch.float32,
    ):
    """Calculates the SVD of the two layers, and alters the weights of the
    layers to be sqrt(S)*U and sqrt(S)*V, and alters the biases of layer_1
    so that it always gives the same output. Only works if there is no
    non-linear activation function between the two layers.

    Args:
        W_in (Tensor): [n_heads, d_head, d_model] OR [d_mid, d_model]
        W_out (Tensor): [d_model, d_mid]
        b_in (Tensor): [n_heads, d_head] OR [d_mid]
        b_out (Tensor): [d_model]
        n_heads (Optional, int): The number of heads to split the [d_mid] into.
        svd_dtype (torch.dtype): dtype to use for SVD

    Output:
        inv_out (InverseLinear): The inverse of the new layer_2
        updated_weights (dict): Dict of updated weights
    """

    # Sort out shapes
    orig_shape_in, orig_shape_out = W_in.shape, W_out.shape
    preshaped = (len(orig_shape_in) == 3)
    dtype, device = W_out.dtype, W_out.device

    if preshaped: # Get the parameters from the MLP
        assert n_heads is None
        (n_heads, d_head, d_model) = orig_shape_in

    else: # Reshape and get parameters
        assert n_heads is not None
        assert orig_shape_in[0] % n_heads == 0, "n_heads must divide d_model"
        d_head  = orig_shape_in[0] // n_heads
        d_model = orig_shape_in[1]

    # pre-compute the effect of layer 1 biases to reconstruct biases later
    # Get layer 1 and layer 2 weights in correct shapes for SVD
    b_in_effect = F.linear(input=b_in.flatten(), weight=W_out, bias=None)
    W_in_heads  = W_in.reshape((n_heads, d_head, d_model))
    W_out_heads = W_out.reshape((d_model, n_heads, d_head))

    # perform SVD using only the (correctly shaped) weights
    svd_two_layer_raw_heads(W_in_heads, W_out_heads, svd_dtype)

    # Get new weights in original format
    W_in  = W_in_heads.reshape(orig_shape_in)
    W_out = W_out_heads.reshape(orig_shape_out)

    # Use inverse linear to reconstruct new biases for W_in
    inv_out = InverseLinear(
            original_weights=W_out, original_biases=b_out, n_heads=n_heads,
        ).to(dtype=dtype).to(device)

    b_in = inv_out.fc(b_in_effect).reshape(orig_shape_in[:-1])

    # Package together new parameters into dict
    updated_weights = {
        "W_in": W_in, "W_out": W_out, "b_in": b_in, "b_out": b_out,
    }

    return inv_out, updated_weights

def mlp_svd_two_layer(
        layer_1: torch.nn.Linear,
        layer_2: torch.nn.Linear,
        n_heads: int = 1,
        svd_dtype: torch.dtype = torch.float32,
    ):
    """Calculates the SVD of the two layers, and alters the weights of the
    layers to be sqrt(S)*U and sqrt(S)*V, and alters the biases of layer_1
    so that it always gives the same output. Only works if there is no
    activation function between the two layers.

    Args:
        layer_1 (torch.nn.Linear): The first layer of the MLP
        layer_2 (torch.nn.Linear): The second layer of the MLP
        d_head (float): The dimension of the head
        svd_dtype (torch.dtype, optional): The dtype to use for the SVD.
            Defaults to torch.float32.

    Returns:
        inv_out (InverseLinear): The inverse of the new layer_2
    """
    params_1 = layer_1.state_dict()
    params_2 = layer_2.state_dict()
    W_in,  B_in  = layer_1.weight, layer_1.bias
    W_out, B_out = layer_2.weight, layer_2.bias

    inv_out, updated_weights = mlp_svd_two_layer_raw(
        W_in, W_out, B_in, B_out, n_heads, svd_dtype )

    params_1["weight"] = updated_weights["W_in"]
    params_1["bias"]   = updated_weights["b_in"]
    layer_1.load_state_dict(params_1)

    params_2["weight"] = updated_weights["W_out"]
    params_2["bias"]   = updated_weights["b_out"]
    layer_2.load_state_dict(params_2)

    return inv_out



"""Hooked Transformer RMS Norm Pre Component.

This module contains all the component :class:`RMSNormPre`.
"""
from typing import Dict, Union

import torch
import torch.nn as nn
from jaxtyping import Float

from transformer_lens.hook_points import HookPoint
from transformer_lens.HookedTransformerConfig import HookedTransformerConfig


class RMSNormPre(nn.Module):
    def __init__(self, cfg: Union[Dict, HookedTransformerConfig]):
        """RMSNormPre - LayerNormPre without the centering and bias (RMS = Root Mean Square)"""
        super().__init__()
        self.cfg = HookedTransformerConfig.unwrap(cfg)
        self.eps = self.cfg.eps

        # Adds a hook point for the normalisation scale factor
        self.hook_scale = HookPoint()  # [batch, pos]
        self.hook_normalized = HookPoint()  # [batch, pos, length]

    def forward(
        self, x: Float[torch.Tensor, "batch pos length"]
    ) -> Float[torch.Tensor, "batch pos length"]:
        if self.cfg.dtype not in [torch.float32, torch.float64]:
            x = x.to(torch.float32)

        scale: Float[torch.Tensor, "batch pos 1"] = self.hook_scale(
            (x.pow(2).mean(-1, keepdim=True) + self.eps).sqrt()
        )
        return self.hook_normalized(x / scale).to(self.cfg.dtype)  # [batch, pos, length]



"""MLP Factory

Centralized location for creating any MLP needed within TransformerLens
"""
from transformer_lens.components.mlps.can_be_used_as_mlp import CanBeUsedAsMLP
from transformer_lens.components.mlps.gated_mlp import GatedMLP
from transformer_lens.components.mlps.gated_mlp_4bit import GatedMLP4Bit
from transformer_lens.components.mlps.mlp import MLP
from transformer_lens.components.mlps.moe import MoE
from transformer_lens.HookedTransformerConfig import HookedTransformerConfig


class MLPFactory:
    @staticmethod
    def create_mlp(cfg: HookedTransformerConfig) -> CanBeUsedAsMLP:
        if cfg.num_experts:
            return MoE(cfg)
        elif cfg.gated_mlp:
            return GatedMLP(cfg) if not cfg.load_in_4bit else GatedMLP4Bit(cfg)
        else:
            return MLP(cfg)



""" This file cointains hook modules which are attached to the model.
"""

from typing import Dict, List
import torch
from torch import Tensor
import warnings

######################################################################################
# Define Hooks on Neuron Activations we can use. EG: Neuron Mask Class
######################################################################################

# Model way of storing hooks that are currently in use
class HookConfig:
    def __init__(self, n_layers=None):
        self.hook_points = {
            "pre_attn": {},
            "attn_pre_out": {},
            "post_attn": {},
            "pre_mlp": {},
            "mlp_pre_out": {},
            "post_mlp": {},
            "pre_decoder": {},
            "post_decoder": {},
        }
        self.n_layers = n_layers

    def from_string(self, config_string):
        for line in config_string.strip().split('\n'):
            parts = line.split(':')
            if len(parts) == 2:
                point, hooks = parts
                layer = None
            elif len(parts) == 3:
                point, layer, hooks = parts
                layer = int(layer.strip())
            else:
                raise ValueError(f"Invalid config line: {line}")

            point = point.strip()
            hooks = hooks.strip()
            if hooks:  # Only add hooks if the string is not empty
                for hook in hooks.split(','):
                    self.add_hook(point, hook.strip(), layer)
        return self

    def add_hook(self, point, hook_type, layer=None):
        if point not in self.hook_points:
            raise ValueError(f"Invalid hook point: {point}")
        if layer is None:
            if 'all' not in self.hook_points[point]:
                self.hook_points[point]['all'] = []
            self.hook_points[point]['all'].append(hook_type)
        else:
            if layer not in self.hook_points[point]:
                self.hook_points[point][layer] = []
            self.hook_points[point][layer].append(hook_type)

    def __str__(self):
        result = []
        for point, layers in self.hook_points.items():
            for layer, hooks in layers.items():
                if hooks:
                    if layer == 'all':
                        result.append(f"{point}: {', '.join(hooks)}")
                    else:
                        result.append(f"{point}: {layer}: {', '.join(hooks)}")
        return '\n'.join(result)

    def get_hooks(self, point, layer):
        hooks = self.hook_points[point].get('all', [])
        hooks += self.hook_points[point].get(layer, [])
        return hooks

class HookMap:
    """Class that holds all hooks."""
    def __init__(self, hook_config):
        self.hook_config: HookConfig = hook_config
        if isinstance(hook_config, str):
            self.hook_config = HookConfig().from_string(hook_config)
        self.collects = {}
        self.neuron_masks = {}
        self.neuron_actadds = {}
        self.neuron_postbiases = {}
        self.neuron_offsets = {}
        self.neuron_unoffsets = {}
        self.neuron_replace = {}
        self.neuron_whiten = {}
        self.neuron_unwhiten = {}
        self.neuron_sae_encode = {}
        self.neuron_sae_decode = {}
        self.handles: list = []

    def __str__(self):
        attributes = []
        for attr, value in self.__dict__.items():
            attributes.append(f"{attr}: {value}")

        return f"ActiveHooks({', '.join(attributes)})"

    def __getitem__(self, component: str):
        return HookMapComponent(self, component)

    @property
    def hooks_raw(self):
        return {
            "collect":  self.collects,
            "mask":     self.neuron_masks,
            "actadd":   self.neuron_actadds,
            "postbias": self.neuron_postbiases,
            "offset":   self.neuron_offsets,
            "unoffset": self.neuron_offsets,
            "replace":  self.neuron_replace,
            "whiten":   self.neuron_whiten,
            "unwhiten": self.neuron_unwhiten,
            "sae_encode": self.neuron_sae_encode,
            "sae_decode": self.neuron_sae_decode,
        }

    @property
    def all_hooks(self):
        hook_list = []
        for _hook_type_dict in self.hooks_raw.values():
            for _location, hook in _hook_type_dict.items():
                hook_list.append(hook)
        return hook_list

    def get_hook_fn(self, hook_type, name, activation, device, dtype):
        _hooks = self.hooks_raw[hook_type]
        if name not in _hooks:
            if hook_type == "collect":
                _hooks[name] = NeuronSave()
            elif hook_type == "mask":
                _hooks[name] = NeuronMask(activation.shape[2:]).to(device, dtype)
            elif hook_type == "actadd":
                _hooks[name] = NeuronActAdd(device, dtype)
            elif hook_type == "postbias":
                _hooks[name] = NeuronPostBias(activation.shape[2:]).to(device, dtype)
            elif hook_type == "offset":
                _hooks[name] = NeuronOffset(activation.shape[2:]).to(device, dtype)
            elif hook_type == "unoffset":
                assert name in self.neuron_offsets
                _hooks[name] = self.neuron_offsets[name]
            elif hook_type == "replace":
                _hooks[name] = NeuronReplace(device, dtype)
            elif hook_type == "whiten":
                _hooks[name] = NeuronWhiten(activation.shape[2:]).to(device, dtype)
            elif hook_type == "unwhiten":
                assert name in self.neuron_whiten
                _hooks[name] = self.neuron_whiten[name]
            elif hook_type == "sae_encode":
                _hooks[name] = NeuronSAE(device, dtype)
            elif hook_type == "sae_decode":
                assert name in self.neuron_sae_encode
                _hooks[name] = self.neuron_sae_encode[name]

        curr_hook = _hooks[name]
        if hook_type == "unoffset":
            return curr_hook.undo
        if hook_type == "unwhiten":
            return curr_hook.undo
        if hook_type == "sae_encode":
            return curr_hook.encode
        if hook_type == "sae_decode":
            return curr_hook.decode
        return curr_hook

    # Generic Helper functions for get/set hook data
    def get_data(self, name=None, data_type=None):
        if data_type == "collect":
            # collect is desctructively gotten (to save memory)
            if name in self.collects:
                data = self.collects[name].activation
                self.collects[name].activation = None
                return data
            return None
        elif data_type in self.hooks_raw:
            return self.hooks_raw[data_type][name]
        else:
            raise ValueError(f"Unknown data type: {data_type}")

    def set_hook_parameter(self, name, param_type, value):
        if param_type == "mask":
            self.neuron_masks[name].set_mask(value)
        elif param_type == "actadd":
            self.neuron_actadds[name].set_actadd(value)
        elif param_type == "postbias":
            self.neuron_postbiases[name].param.data = value
        else:
            raise ValueError(f"Unknown parameter type: {param_type}")

    def get_layer_names(self, component, layers=None):
        layers = range(self.hook_config.n_layers) if layers is None else layers
        return [f"layer_{i}_{component}" for i in layers]

    def get_all_layer_data(self, component, data_type, layers=None):
        layer_names = self.get_layer_names(component, layers)
        return [self.get_data(name, data_type) for name in layer_names]

    def set_all_layer_parameters(self, component, param_type, values, layers=None):
        layer_names = self.get_layer_names(component, layers)
        for name, value in zip(layer_names, values):
            self.set_hook_parameter(name, param_type, value)

    # Methods for specific hook types
    def get_all_layer_activations(self, component: str, layers: List[int] | None =None):
        layer_names = self.get_layer_names(component, layers)
        return [self.get_data(name, "collect") for name in layer_names]

    def disable_all_collect_hooks(self):
        for name, hook in self.collects.items():
            hook.enabled = False

    def enable_collect_hooks(self, components=None, layers=None, run_assert=False):
        if components is None:
            components = self.hook_config.hook_points.keys()
        if isinstance(components, str):
            components = [components]
        if layers is None:
            layers = range(self.hook_config.n_layers)
        if isinstance(layers, int):
            layers = [layers]

        for component in components:
            for layer in layers:
                hook_name = f"layer_{layer}_{component}"
                if run_assert:
                    assert hook_name in self.collects
                if hook_name in self.collects:
                    self.collects[hook_name].enabled = True

    def delete_mlp_neurons(self, remove_indices, layer: int = None):
        return self["mlp_pre_out"].delete_neurons(remove_indices, layer)

    def delete_attn_neurons(self, remove_indices, layer: int = None):
        return self["attn_pre_out"].delete_neurons(remove_indices, layer)

    def reset_neuron_replace(self):
        [h.reset() for h in self.neuron_replace.values()]

    def reset(self):
        [h.reset() for h in self.all_hooks]

class HookMapComponent:
    def __init__(self, hooks: HookMap, component: str):
        self.hooks: HookMap = hooks
        self.component: str = component

    def __getitem__(self, data_type):
        layers = None
        if isinstance(data_type, tuple):
            data_type, layer = data_type
            layers = [layer]
        if data_type == "collect":
            data = self.hooks.get_all_layer_activations(self.component, layers)
        elif data_type in ["mask", "actadd", "postbias", "offset"]:
            data = self.hooks.get_all_layer_data(self.component, data_type, layers)
        else:
            raise ValueError(f"Unknown data type: {data_type}")
        if len(data) == 1:
            return data[0]
        return data

    def __setitem__(self, data_type, value):
        if data_type in ["mask", "actadd", "postbias", "offset"]:
            self.hooks.set_all_layer_parameters(self.component, data_type, value)
        else:
            raise ValueError(f"Cannot set data type: {data_type}")

    # Hook-specific functions
    def delete_neurons(self, remove_indices, layer: int = None):
        def delete_layer_neurons(nn_mask: NeuronMask, rm_idx):
            device, dtype = nn_mask.param.device, bool
            rm_idx = rm_idx.to(device, dtype).reshape(nn_mask.param.shape)
            keep_indices = torch.logical_not(rm_idx)
            nn_mask.delete_neurons(keep_indices)
        nn_masks = self["mask"]
        if layer is not None:
            return delete_layer_neurons(nn_masks[layer], remove_indices)
        for layer, nn_mask in enumerate(nn_masks):
            delete_layer_neurons(nn_mask, remove_indices[layer])

    def set_offsets(self, offset_val, layer:int = None):
        def set_layer_offset(nn_offset, val):
            device, dtype = nn_offset.param.device, nn_offset.param.dtype
            val = val.to(device, dtype).reshape(nn_offset.shape)
            nn_offset.set_offset(val)
        neuron_offsets = self["offset"]
        if layer is not None:
            return set_layer_offset(neuron_offsets[layer], offset_val)
        for layer, neuron_offset in enumerate(neuron_offsets):
            set_layer_offset(neuron_offset, offset_val[layer])

#####################################################################################
# Define Hook Classes
#####################################################################################

class NeuronFunctionList(torch.nn.Module):
    """ Class for storing all the Neuron Masks"""

    def __init__(self, neuron_function_list):
        super(NeuronFunctionList, self).__init__()
        # list all the Neuron Masks as a torch accessible list of parameters
        self.layers = torch.nn.ModuleList(neuron_function_list)

    def forward(self, x):
        "Given [layer, activation], returns all the activations masked for each layer."
        y = []
        for act in x:
            y.append(act)
        return torch.stack(y)

    def __getitem__(self, index: int):
        return self.layers[index]

class NeuronSave(torch.nn.Module):
    """Class that saves activations to self"""

    def __init__(self):
        super().__init__()
        self.activation = None
        self.enabled = False
        self.concat_mode = False

    def forward(self, x: Tensor):
        if self.enabled and self.concat_mode and self.activation is not None:
            self.activation = torch.concat([self.activation, x], dim=1) # batch token *dims
        elif self.enabled:
            self.activation = x
        return x

    def reset(self):
        self.activation = None

# Neuron Mask. EG: [a, b, c] -> [a, 0, c]
class NeuronMask(torch.nn.Module):
    """Class for creating a mask for a single layer of a neural network."""

    def __init__(self, shape, act_fn: str = "step"):
        super(NeuronMask, self).__init__()
        self.act_fn = act_fn
        self.shape: torch.Size = None
        self.param: torch.nn.Parameter = None
        self.offset: torch.nn.Parameter = None
        self.reinit_hook(shape=shape)

    def check_shapes_match(self, x):
        curr_shape  = torch.Size(self.shape)
        input_shape = torch.Size(x.shape[-len(curr_shape):])
        return curr_shape == input_shape, f"{curr_shape} vs {input_shape} (from {x.shape})"

    def reinit_hook(self, x=None, shape=None):
        # batch, token, (d_model or otherwise)
        if x is not None:
            new_shape, new_dtype = x.shape[2:], x.dtype
        elif shape is not None:
            new_shape, new_dtype = shape, torch.float32
        else:
            raise ValueError("Either x or shape must be provided to init NeuronMask")

        self.shape = new_shape
        vec = torch.ones(new_shape, dtype=new_dtype)
        # initialize mask as nn.Parameter of ones
        if self.act_fn == "sigmoid":
            vec[...] = torch.inf
        self.param = torch.nn.Parameter(vec)
        self.offset = torch.nn.Parameter(torch.zeros_like(vec))

    def get_mask(self):
        # if step, we want heaviside step function. ie: mask = mask > 0
        if self.act_fn == "step":
            return (self.param > 0)
        if self.act_fn == "sigmoid":
            return torch.sigmoid(self.param)
        if self.act_fn == "tanh":
            return torch.tanh(self.param)
        if self.act_fn == "relu":
            return torch.relu(self.param)
        if callable(self.act_fn):
            return self.act_fn(self.param)
        raise ValueError(f"Unknown activation function: {self.act_fn}")

    def get_offset(self, x):
        mask = self.get_mask().to(x.dtype)
        inv_mask = 1 - mask
        offset = self.offset * inv_mask
        return offset

    def set_mask(self, new_mask: Tensor):
        params: Dict[str, Tensor] = self.state_dict()
        params["param"] = new_mask.view(self.shape)
        self.load_state_dict(params)

    def set_offset(self, offset: Tensor):
        params: Dict[str, Tensor] = self.state_dict()
        params["offset"] = offset.view(self.shape)
        self.load_state_dict(params)

    def delete_neurons(self, keep_indices: Tensor):
        params: Dict[str, Tensor] = self.state_dict()
        params["param"] = params["param"] * keep_indices.to(params["param"].device)
        self.load_state_dict(params)

    def get_inverse_mask(self, x):
        mask = self.get_mask().to(x.dtype)
        return 1 - mask

    def inverse_mask(self, x, offset=False):
        inv_mask = self.get_inverse_mask(x)
        # TODO: allow inverse mask to work with offset. Not sure when needed though.
        assert offset == False
        return x * inv_mask

    def forward(self, x):
        is_match, msg = self.check_shapes_match(x)
        if not is_match:
            print(f"Shape mismatch: {msg}, reinitialising mask hook")
            self.reinit_hook(x)
        self.to(x.device)
        mask = self.get_mask()
        offset = self.get_offset(x)
        return x * mask + offset

    def __str__(self):
        return f"""NeuronMask(
        act_fn: {self.act_fn}
        param: {self.param}
        offset: {self.offset}
        )"""

    def reset(self):
        self.param.data = torch.ones_like(self.param)
        self.offset.data = torch.zeros_like(self.offset)

# Positional Neuron Activation Addition.
class NeuronActAdd(torch.nn.Module):
    """ MVP for Position Dependant Neuron Activation Offseting.
    Inspired by ActAdd paper.
    # EG: [[a], [b], [c], [d], ...] -> [[a+x], [b'+y], [c'+z], [d'], ...]
    # where a,b,c,d are the "real activations", x,y,z are the added vectors.

    the "autorestart" works by assuming the first input is of size > 1 and is cached.
    """
    def __init__(self, device, dtype, autorestart: bool=True):
        super(NeuronActAdd, self).__init__()
        self.device = device
        self.dtype = dtype

        # leave as uninitialised initially
        self.shape = [0]
        self.param = torch.nn.Parameter(torch.zeros(self.shape))
        self.max_tokens = 0

        self.tokens_seen = 0
        self.autorestart = autorestart

    def restart(self):
        self.tokens_seen = 0

    def reset(self):
        self.param = torch.nn.Parameter(torch.zeros(self.shape))
        self.max_tokens = 0
        self.restart()

    def set_actadd(self, offset: Tensor):
        self.shape = offset.shape
        self.max_tokens = self.shape[0]
        self.param = torch.nn.Parameter(
            offset.to(device=self.device, dtype=self.dtype)
        )
        self.restart()

    def to(self, device=None, dtype=None, *args, **kwargs):
        super(NeuronActAdd, self).to(device, dtype, *args, **kwargs)
        if not device is None:
            self.device = device
        if not dtype is None:
            self.dtype = dtype

    def forward(self, x):
        # load input vector (do not modify in place)
        x = x.clone()
        n_new_tokens = x.shape[1]
        # autoreset assuming cache goes like [0,1,2,3], [4], [5], ...
        if self.autorestart and n_new_tokens > 1:
            self.restart()
        # if we used up all the add act vectors we can skip
        if self.tokens_seen >= self.max_tokens:
            return x
        # othewise, do the act add stuff with remaining vectors left
        tokens_left = self.max_tokens - self.tokens_seen
        n_tokens    = min([tokens_left, n_new_tokens])

        x[:, :n_tokens]     += self.param[ self.tokens_seen:self.tokens_seen+n_tokens]
        self.tokens_seen += n_tokens
        return x

class NeuronReplace(torch.nn.Module):
    """ Replace neuron activations at specific token indices
    # EG: [[a], [b], [c], [d], ...] -> [[a], [b], [X], [d], ...]

    the "autorestart" works by assuming the first input is of size > 1 and is cached.
    """
    def __init__(self, device, dtype, autorestart: bool=True):
        super(NeuronReplace, self).__init__()
        self.device = device
        self.dtype = dtype

        # Use a ParameterDict to store parameters
        self.param = torch.nn.ParameterDict()
        self.max_tokens = 0

        self.tokens_seen = 0
        self.autorestart = autorestart

    def restart(self):
        self.tokens_seen = 0

    def reset(self):
        self.max_tokens = 0
        self.param.clear()
        self.restart()

    def add_token(self, token_index, value):
        # Convert the value to a Parameter before adding
        self.param[str(token_index)] = torch.nn.Parameter(value.to(self.device, self.dtype))
        self.max_tokens = max([self.max_tokens, token_index+1])

    def to(self, device=None, dtype=None, *args, **kwargs):
        super(NeuronReplace, self).to(device, dtype, *args, **kwargs)
        if device is not None:
            self.device = device
        if dtype is not None:
            self.dtype = dtype

    def forward(self, x):
        # load input vector (do not modify in place)
        x = x.clone()
        n_new_tokens = x.shape[1]
        # autorestart assuming cache goes like [0,1,2,3], [4], [5], ...
        if self.autorestart and n_new_tokens > 1:
            self.restart()
        # if we used up all the add act vectors we can skip
        if self.tokens_seen >= self.max_tokens:
            return x

        # othewise, do the neuron replacement stuff with remaining vectors left
        tokens_left = self.max_tokens - self.tokens_seen
        n_tokens    = min([tokens_left, n_new_tokens])
        for local_index in range(n_tokens):
            token_index = local_index + self.tokens_seen
            if str(token_index) in self.param:
                x[:, local_index] = self.param[str(token_index)].reshape(x.shape[2:])

        self.tokens_seen += n_tokens
        return x

class NeuronOffset(torch.nn.Module):
    def __init__(self, shape):
        super(NeuronOffset, self).__init__()
        self.shape = shape
        _vec = torch.zeros(shape, dtype=torch.float32)
        self.param = torch.nn.Parameter(_vec)

    def set_offset(self, offset: Tensor):
        params: Dict[str, Tensor] = self.state_dict()
        params["param"] = offset.view(self.shape)
        self.load_state_dict(params)

    def forward(self, x):
        self.to(x.device)
        return x + self.param

    def undo(self, x):
        return x - self.param

    def reset(self):
        self.param.data = torch.zeros_like(self.param)

# Neuron Post Bias (EG: For SVD and stuff) out -> out + bias
class NeuronPostBias(torch.nn.Module):
    """Container for holding after-the-fact biases in the model."""

    def __init__(self, shape):
        super(NeuronPostBias, self).__init__()
        self.shape = shape
        _vec = torch.zeros(shape, dtype=torch.float32)
        self.param = torch.nn.Parameter(_vec)

    def get_bias(self, x):
        shape = x.shape
        bias  = self.param
        if self.shape == shape:
            return bias
        try:
            bias = bias.view(shape[-1]) # normal shape
        except:
            bias = bias.view(-1, shape[-1]) # multi head attention shape
        return bias

    def forward(self, x):
        return x + self.get_bias(x)

    def reset(self):
        self.param.data = torch.zeros_like(self.param)

class NeuronWhiten(torch.nn.Module):
    def __init__(self, shape):
        super(NeuronWhiten, self).__init__()
        self.shape = shape
        self.d_model = self.shape.numel()
        self.offset   = torch.nn.Parameter(torch.zeros(self.shape))
        self.rotate   = torch.nn.Linear(self.d_model, self.d_model, bias=False)
        self.scale    = torch.nn.Parameter(torch.ones(self.d_model))
        self.rotate_inv = torch.nn.Linear(self.d_model, self.d_model, bias=False)
        self.reset()

    def reset(self):
        self.offset.data = torch.zeros(self.shape)
        self.rotate.weight.data = torch.diag(torch.ones(self.d_model))
        self.scale.data = torch.ones(self.d_model)
        self.rotate_inv.weight.data = torch.diag(torch.ones(self.d_model))

    @property
    def unscale(self):
        return torch.clamp( self.scale ** -1, max=1e6 )

    def forward(self, x):
        x = x + self.offset
        x = x.reshape([*x.shape[:2], -1])
        x = self.rotate(x)
        x = x * self.scale
        return x

    def undo(self, x):
        x = x * self.unscale
        x = self.rotate_inv(x)
        x = x.reshape(x.shape[:2] + self.shape)
        x = x - self.offset
        return x

class NeuronSAE(torch.nn.Module):
    def __init__(self, device, dtype):
        super(NeuronSAE, self).__init__()
        self.device = device
        self.dtype = dtype
        self.sae = None
        self.sae_config = None

    def load_sae(self, sae, sae_config=None):
        self.sae = sae.to(self.device, self.dtype)
        self.sae_config = sae_config

    def load_sae_from_pretrained(self, release, sae_id):
        try:
            from sae_lens import SAE
        except ImportError:
            raise ImportError("sae_lens not installed. Please install it with `pip install sae-lens`.")

        sae, cfg_dict, sparsity = SAE.from_pretrained(
            release = release,
            sae_id =  sae_id,
            device = self.device,
        )
        self.load_sae(sae, cfg_dict)

    def forward(self, x):
        return self.encode(x)

    def encode(self, x):
        if self.sae is None:
            warnings.warn("SAE not loaded. Call load_sae() first.")
            return x
        return self.sae.encode(x)

    def decode(self, x):
        if self.sae is None:
            return x
        return self.sae.decode(x)

    def reset(self):
        self.sae = None
        self.sae_config = None



"""Addmm

Implementations of Addmm functions matching Huggingface implementations.
"""
import torch
from jaxtyping import Float


def vanilla_addmm(
    input: Float[torch.Tensor, "... #o"],  # Must be broadcastable to "m o"
    mat1: Float[torch.Tensor, "m n"],
    mat2: Float[torch.Tensor, "n o"],
) -> Float[torch.Tensor, "m o"]:
    """Typechecked version of torch.addmm.

    Note that both mat1 and mat2 *must* be 2d matrices.
    """
    return torch.addmm(input, mat1, mat2)


def batch_addmm(
    bias: Float[torch.Tensor, "... #d_out"],  # Must be broadcastable to "... d_out"
    weight: Float[torch.Tensor, "d_in d_out"],
    x: Float[torch.Tensor, "... d_in"],
) -> Float[torch.Tensor, "... d_out"]:
    """Fused add-multiply with support for batch dimensions.

    Must match the Huggingface Conv1D implementation exactly.
    https://github.com/huggingface/transformers/blob/9ba9369a2557e53a01378199a9839ec6e82d8bc7/src/transformers/pytorch_utils.py#L102-L106
    """
    n_output_features = weight.shape[-1]
    size_out = x.size()[:-1] + (n_output_features,)
    x = vanilla_addmm(bias, x.view(-1, x.size(-1)), weight)
    x = x.view(size_out)
    return x



"""Streamlined Neo-Taker Model with TransformerLens compatibility."""

from typing import List, Optional, Union
import warnings
import torch
import torch.nn as nn
from torch import Tensor
from transformers import AutoTokenizer, AutoModelForCausalLM, PreTrainedModel, AutoConfig

from .data_classes import DtypeMap
from .hooks import HookedRootModule, HookPoint
from .component_maps import MapConfigClass, ModelMap, convert_hf_model_config

class Model(HookedRootModule):
    """Neo-Taker model with TransformerLens-style hooks."""
    
    def __init__(
        self,
        model_repo: str = "nickypro/tinyllama-15m",
        model_device: str = None,
        dtype: str = "bf16",
        tokenizer_repo: str = None,
        eval_mode: bool = True,
        model_kwargs: dict = None,
    ):
        super().__init__()
        
        # Core configuration
        self.model_repo = model_repo
        self.tokenizer_repo = tokenizer_repo or model_repo
        self.eval_mode = eval_mode
        self.model_kwargs = model_kwargs or {}
        
        # Dtype handling
        self.dtype_map = DtypeMap(dtype)
        self.dtype = self.dtype_map._dtype
        self.dtype_args = self.dtype_map._dtype_args
        
        # Model components
        self.hf_config: AutoConfig = None
        self.cfg: MapConfigClass = None
        self.tokenizer: AutoTokenizer = None
        self.predictor: AutoModelForCausalLM = None
        self.model: PreTrainedModel = None
        self.map: ModelMap = None
        self.layers: list = None
        # Track which TL-style hook points are actually wired into the forward pass
        self._wired_hook_names: set[str] = set()

        # Initialize the model
        self._init_model()
        self.cfg.device = model_device or ("cuda" if torch.cuda.is_available() else "cpu")
        self._add_hook_points()
        self.setup()  # Setup TransformerLens hooks
        
        if self.eval_mode and self.predictor:
            self.predictor.eval()

    @classmethod
    def from_pretrained(cls, *args, **kwargs):
        """Create a Model instance from a pretrained model. 
        Added for compatibility with TransformerLens syntax.
        Use normal model init syntax for type hints: 
        - Model(model_repo="nickypro/tinyllama-15m", model_device="cpu")
        - Model.from_pretrained("nickypro/tinyllama-15m", model_device="cpu")
        """
        return cls(*args, **kwargs)
    
    def _init_model(self):
        """Initialize the model components."""
        # Load configuration
        self.cfg, self.hf_config = convert_hf_model_config(self.model_repo)
        
        # Load tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.tokenizer_repo, 
            legacy=False, 
            padding_side='left'
        )
        if self.tokenizer.pad_token is None:
            self.tokenizer.pad_token = self.tokenizer.eos_token
        
        # Load model
        model_args = {**self.dtype_args, **self.model_kwargs}
        self.predictor = AutoModelForCausalLM.from_pretrained(
            self.model_repo,
            device_map="auto",
            **model_args,
        )
        
        # Create model mapping
        self.map = ModelMap(self.predictor, self.cfg)
        self.model = self.map["model"]
        self.layers = self.map.layers
        
        print(f"Loaded model '{self.model_repo}' with {self.dtype_map.str_dtype}")
    
    def _add_hook_points(self):
        """Add TransformerLens-style hook points and integrate them into model components."""
        # Embedding hook
        self.hook_embed = HookPoint()
        
        # Layer-specific hooks
        for i in range(self.cfg.n_layers):
            layer = self.layers[i]
            
            # Residual stream hooks
            setattr(self, f'blocks.{i}.hook_resid_pre', HookPoint())
            setattr(self, f'blocks.{i}.hook_resid_mid', HookPoint())
            setattr(self, f'blocks.{i}.hook_resid_post', HookPoint())
            
            # Layer norm hooks (monitoring scale and normalized output)
            setattr(self, f'blocks.{i}.ln1.hook_scale', HookPoint())
            setattr(self, f'blocks.{i}.ln1.hook_normalized', HookPoint())
            setattr(self, f'blocks.{i}.ln2.hook_scale', HookPoint())
            setattr(self, f'blocks.{i}.ln2.hook_normalized', HookPoint())
            
            # Attention hooks (monitoring intermediate values)
            setattr(self, f'blocks.{i}.attn.hook_k', HookPoint())       # After K projection
            setattr(self, f'blocks.{i}.attn.hook_q', HookPoint())       # After Q projection  
            setattr(self, f'blocks.{i}.attn.hook_v', HookPoint())       # After V projection
            setattr(self, f'blocks.{i}.attn.hook_z', HookPoint())       # After attention (pre-output proj)
            setattr(self, f'blocks.{i}.attn.hook_attn_scores', HookPoint())  # Raw attention scores
            setattr(self, f'blocks.{i}.attn.hook_pattern', HookPoint())      # Attention weights (post-softmax)
            setattr(self, f'blocks.{i}.attn.hook_result', HookPoint())       # Attention output per head
            setattr(self, f'blocks.{i}.attn.hook_rot_k', HookPoint())   # After rotary embedding on K
            setattr(self, f'blocks.{i}.attn.hook_rot_q', HookPoint())   # After rotary embedding on Q
            
            # MLP hooks (monitoring inputs/outputs)
            setattr(self, f'blocks.{i}.mlp.hook_pre', HookPoint())           # MLP input
            setattr(self, f'blocks.{i}.mlp.hook_pre_linear', HookPoint())    # After first linear layer
            setattr(self, f'blocks.{i}.mlp.hook_post', HookPoint())          # MLP output
            
            # Input/output hooks (monitoring what goes into/out of components)
            setattr(self, f'blocks.{i}.hook_attn_in', HookPoint())      # Input to attention
            setattr(self, f'blocks.{i}.hook_q_input', HookPoint())      # Input to Q projection
            setattr(self, f'blocks.{i}.hook_k_input', HookPoint())      # Input to K projection
            setattr(self, f'blocks.{i}.hook_v_input', HookPoint())      # Input to V projection
            setattr(self, f'blocks.{i}.hook_mlp_in', HookPoint())       # Input to MLP
            setattr(self, f'blocks.{i}.hook_attn_out', HookPoint())     # Output from attention
            setattr(self, f'blocks.{i}.hook_mlp_out', HookPoint())      # Output from MLP
        
        # Final layer norm hooks
        setattr(self, 'ln_final.hook_scale', HookPoint())
        setattr(self, 'ln_final.hook_normalized', HookPoint())
        
        # Now integrate hooks into the actual model forward pass
        self._integrate_hooks_into_model()
    
    def _integrate_hooks_into_model(self):
        """Integrate hook points using PyTorch's native hook system."""
        
        # Register PyTorch hooks that call our TransformerLens-style hooks
        
        # Hook the embedding layer
        embed_module = self.map["embed"]
        embed_module.register_forward_hook(self._create_pytorch_hook("hook_embed"))
        self._wired_hook_names.add("hook_embed")
        
        # Hook each transformer layer
        for i in range(self.cfg.n_layers):
            self._register_layer_hooks(i)
        
        # Hook final layer norm if it exists
        if "ln_final" in self.map.key_map:
            ln_final = self.map["ln_final"]
            # Hook input (scale) and output (normalized)
            ln_final.register_forward_pre_hook(self._create_pytorch_pre_hook(f"ln_final.hook_scale"))
            ln_final.register_forward_hook(self._create_pytorch_hook(f"ln_final.hook_normalized"))
            self._wired_hook_names.add("ln_final.hook_scale")
            self._wired_hook_names.add("ln_final.hook_normalized")
    
    def _create_pytorch_hook(self, hook_name):
        """Create a PyTorch forward hook that calls our TransformerLens hook."""
        def pytorch_hook(module, input, output):
            # Get our hook point
            hook_point = getattr(self, hook_name)
            # Call it with the output (TransformerLens style)
            return hook_point(output)
        return pytorch_hook
    
    def _create_pytorch_pre_hook(self, hook_name):
        """Create a PyTorch pre-forward hook that calls our TransformerLens hook."""
        def pytorch_pre_hook(module, input):
            # Get our hook point
            hook_point = getattr(self, hook_name)
            # Call it with the input (TransformerLens style)
            if isinstance(input, tuple) and len(input) > 0:
                modified_input = hook_point(input[0])
                if len(input) > 1:
                    return (modified_input,) + input[1:]
                else:
                    return (modified_input,)
            elif isinstance(input, Tensor):
                return hook_point(input)
            else:
                # If input is not a tensor or tuple, just return it unchanged
                return input
        return pytorch_pre_hook
    
    def _register_layer_hooks(self, layer_idx):
        """Register PyTorch hooks for a specific transformer layer."""
        layer = self.layers[layer_idx]
        
        # Get the actual HuggingFace transformer layer
        hf_layer = layer.layer
        
        # Hook residual stream at layer level
        hf_layer.register_forward_pre_hook(self._create_pytorch_pre_hook(f"blocks.{layer_idx}.hook_resid_pre"))
        def resid_post_hook(module, inp, out):
            hp = getattr(self, f"blocks.{layer_idx}.hook_resid_post")
            if isinstance(out, tuple) and len(out) > 0:
                t0 = hp(out[0])
                return (t0,) + out[1:]
            return hp(out)
        hf_layer.register_forward_hook(resid_post_hook)
        self._wired_hook_names.add(f"blocks.{layer_idx}.hook_resid_pre")
        self._wired_hook_names.add(f"blocks.{layer_idx}.hook_resid_post")
        
        # Try to hook attention and MLP components if they exist
        try:
            # Hook attention components (support both LLaMA `self_attn` and GPT-2 `attn`)
            if hasattr(hf_layer, 'self_attn') or hasattr(hf_layer, 'attn'):
                attn = hf_layer.self_attn if hasattr(hf_layer, 'self_attn') else hf_layer.attn

                # Hook attention input
                attn.register_forward_pre_hook(self._create_pytorch_pre_hook(f"blocks.{layer_idx}.hook_attn_in"))

                # For attn_out, the module often returns a tuple; pass only tensor[0] to TL hook
                def attn_out_hook(module, inp, out):
                    hp = getattr(self, f"blocks.{layer_idx}.hook_attn_out")
                    if isinstance(out, tuple) and len(out) > 0:
                        t0 = hp(out[0])
                        return (t0,) + out[1:]
                    return hp(out)
                attn.register_forward_hook(attn_out_hook)
                self._wired_hook_names.add(f"blocks.{layer_idx}.hook_attn_in")
                self._wired_hook_names.add(f"blocks.{layer_idx}.hook_attn_out")

                # Helper to compute head counts
                def _num_heads(which: str) -> int:
                    n_qo_heads = self.cfg.n_heads
                    n_kv_heads = getattr(self.cfg, 'n_key_value_heads', None) or n_qo_heads
                    return n_kv_heads if which in ('k', 'v') else n_qo_heads

                d_head = self.cfg.d_head

                # Custom forward hook for Q/K/V to expose [batch, seq, n_heads, d_head] to hooks
                def _qkv_hook(which: str):
                    def _hook(module, inp, out):
                        hp = getattr(self, f"blocks.{layer_idx}.attn.hook_{which}")
                        if not isinstance(out, torch.Tensor) or out.dim() != 3:
                            return hp(out)
                        B, S, D = out.shape
                        # Heads per stream
                        nH_qo = _num_heads('q')
                        nH_kv = _num_heads('k')
                        d_model_qo = nH_qo * d_head
                        d_model_kv = nH_kv * d_head

                        # Case 1: separate projection with expected dimension
                        nH = _num_heads(which)
                        expected = nH * d_head
                        if D == expected:
                            t = out.view(B, S, nH, d_head)
                            t2 = hp(t)
                            if t2 is None:
                                t2 = t
                            return t2.reshape(B, S, expected)

                        # Case 2: combined QKV (e.g., GPT-2 c_attn): split, apply, and re-concat
                        if D == (d_model_qo + d_model_kv + d_model_kv):
                            q_end = d_model_qo
                            k_end = q_end + d_model_kv
                            q_seg = out[..., :q_end]
                            k_seg = out[..., q_end:k_end]
                            v_seg = out[..., k_end:]

                            if which == 'q':
                                t = q_seg.view(B, S, nH_qo, d_head)
                                t2 = hp(t)
                                if t2 is None:
                                    t2 = t
                                q_seg = t2.reshape(B, S, d_model_qo)
                            elif which == 'k':
                                t = k_seg.view(B, S, nH_kv, d_head)
                                t2 = hp(t)
                                if t2 is None:
                                    t2 = t
                                k_seg = t2.reshape(B, S, d_model_kv)
                            elif which == 'v':
                                t = v_seg.view(B, S, nH_kv, d_head)
                                t2 = hp(t)
                                if t2 is None:
                                    t2 = t
                                v_seg = t2.reshape(B, S, d_model_kv)
                            return torch.cat([q_seg, k_seg, v_seg], dim=-1)

                        # Fallback: warn and pass through hook without reshaping
                        warnings.warn(
                            f"Unexpected proj dim at layer {layer_idx} for {which}: {D}; cannot infer head split"
                        )
                        return hp(out)
                    return _hook

                # Hook Q, K, V projections using model mapping
                mod_to_whiches = {}
                proj_modules = {}
                if "q_proj" in layer.key_map:
                    proj_modules['q'] = layer["q_proj"]
                    self._wired_hook_names.add(f"blocks.{layer_idx}.hook_q_input")
                    self._wired_hook_names.add(f"blocks.{layer_idx}.attn.hook_q")
                if "k_proj" in layer.key_map:
                    proj_modules['k'] = layer["k_proj"]
                    self._wired_hook_names.add(f"blocks.{layer_idx}.hook_k_input")
                    self._wired_hook_names.add(f"blocks.{layer_idx}.attn.hook_k")
                if "v_proj" in layer.key_map:
                    proj_modules['v'] = layer["v_proj"]
                    self._wired_hook_names.add(f"blocks.{layer_idx}.hook_v_input")
                    self._wired_hook_names.add(f"blocks.{layer_idx}.attn.hook_v")

                # Group whiches by underlying module (to deduplicate pre-hooks for combined QKV)
                for which, mod in proj_modules.items():
                    mod_to_whiches.setdefault(id(mod), {"module": mod, "whiches": []})["whiches"].append(which)

                # Register a single pre-hook per unique module that chains q/k/v input hooks
                for entry in mod_to_whiches.values():
                    mod = entry["module"]
                    whiches = entry["whiches"]
                    def _chained_prehook(module, input, whiches=whiches):
                        hp_inputs = {
                            'q': getattr(self, f"blocks.{layer_idx}.hook_q_input"),
                            'k': getattr(self, f"blocks.{layer_idx}.hook_k_input"),
                            'v': getattr(self, f"blocks.{layer_idx}.hook_v_input"),
                        }
                        if isinstance(input, tuple) and len(input) > 0:
                            x = input[0]
                            for w in whiches:
                                x2 = hp_inputs[w](x)
                                if x2 is not None:
                                    x = x2
                            return (x,) + input[1:]
                        elif isinstance(input, Tensor):
                            x = input
                            for w in whiches:
                                x2 = hp_inputs[w](x)
                                if x2 is not None:
                                    x = x2
                            return x
                        return input
                    mod.register_forward_pre_hook(_chained_prehook)

                # Register forward hooks (one per which) to expose per-head Q/K/V
                for which, mod in proj_modules.items():
                    mod.register_forward_hook(_qkv_hook(which))

                # Custom pre-hook for Z to expose [batch, seq, n_heads, d_head] to hooks
                def _z_prehook(module, input):
                    hp = getattr(self, f"blocks.{layer_idx}.attn.hook_z")
                    if not (isinstance(input, tuple) and len(input) > 0 and isinstance(input[0], torch.Tensor)):
                        return input
                    x = input[0]
                    if x.dim() != 3:
                        return input
                    B, S, D = x.shape
                    nH = _num_heads('q')  # z uses qo heads
                    expected = nH * d_head
                    if D != expected:
                        warnings.warn(f"Unexpected z dim: {D} != n_heads({nH})*d_head({d_head}) at layer {layer_idx}")
                        t2 = hp(x)
                        return (t2,) + input[1:] if len(input) > 1 else (t2,)
                    t = x.view(B, S, nH, d_head)
                    t2 = hp(t)
                    if t2 is None:
                        t2 = t
                    flat = t2.reshape(B, S, expected)
                    if len(input) > 1:
                        return (flat,) + input[1:]
                    else:
                        return (flat,)

                if "o_proj" in layer.key_map:
                    out_proj = layer["o_proj"]
                    out_proj.register_forward_pre_hook(_z_prehook)
                    out_proj.register_forward_hook(self._create_pytorch_hook(f"blocks.{layer_idx}.hook_attn_out"))
                    self._wired_hook_names.add(f"blocks.{layer_idx}.attn.hook_z")
                    self._wired_hook_names.add(f"blocks.{layer_idx}.hook_attn_out")
            
            # Hook MLP components
            if hasattr(hf_layer, 'mlp'):
                mlp = hf_layer.mlp
                
                # Hook MLP input/output
                mlp.register_forward_pre_hook(self._create_pytorch_pre_hook(f"blocks.{layer_idx}.hook_mlp_in"))
                # Wire resid_mid at MLP input (residual after attention)
                mlp.register_forward_pre_hook(self._create_pytorch_pre_hook(f"blocks.{layer_idx}.hook_resid_mid"))
                mlp.register_forward_hook(self._create_pytorch_hook(f"blocks.{layer_idx}.hook_mlp_out"))
                self._wired_hook_names.add(f"blocks.{layer_idx}.hook_mlp_in")
                self._wired_hook_names.add(f"blocks.{layer_idx}.hook_resid_mid")
                self._wired_hook_names.add(f"blocks.{layer_idx}.hook_mlp_out")
                
                # Hook MLP components
                mlp.register_forward_pre_hook(self._create_pytorch_pre_hook(f"blocks.{layer_idx}.mlp.hook_pre"))
                mlp.register_forward_hook(self._create_pytorch_hook(f"blocks.{layer_idx}.mlp.hook_post"))
                self._wired_hook_names.add(f"blocks.{layer_idx}.mlp.hook_pre")
                self._wired_hook_names.add(f"blocks.{layer_idx}.mlp.hook_post")
            
            # Hook layer norms using model mapping
            # Use TransformerLens naming: ln1 and ln2 via map only
            if "ln1" in layer.key_map:
                ln1 = layer["ln1"]
                ln1.register_forward_pre_hook(self._create_pytorch_pre_hook(f"blocks.{layer_idx}.ln1.hook_scale"))
                ln1.register_forward_hook(self._create_pytorch_hook(f"blocks.{layer_idx}.ln1.hook_normalized"))
                self._wired_hook_names.add(f"blocks.{layer_idx}.ln1.hook_scale")
                self._wired_hook_names.add(f"blocks.{layer_idx}.ln1.hook_normalized")
            if "ln2" in layer.key_map:
                ln2 = layer["ln2"]
                ln2.register_forward_pre_hook(self._create_pytorch_pre_hook(f"blocks.{layer_idx}.ln2.hook_scale"))
                ln2.register_forward_hook(self._create_pytorch_hook(f"blocks.{layer_idx}.ln2.hook_normalized"))
                self._wired_hook_names.add(f"blocks.{layer_idx}.ln2.hook_scale")
                self._wired_hook_names.add(f"blocks.{layer_idx}.ln2.hook_normalized")

        except Exception as e:
            print(f"Warning: Could not register all hooks for layer {layer_idx}: {e}")

    def get_activation_map(self, wired_only: bool = True) -> dict[str, HookPoint]:
        """Return a map of TL-style activation hook names to HookPoint modules.

        If wired_only is True, only returns hook points that are actually integrated into the
        forward pass (excludes non-existent points like attn_scores/pattern where unavailable).
        """
        all_points = self.hook_points()
        if wired_only:
            return {name: hp for name, hp in all_points.items() if name in self._wired_hook_names}
        return all_points

    def list_activation_points(self, wired_only: bool = True) -> List[str]:
        """List available TL-style activation hook names.

        If wired_only is True, only lists those actually connected to real activations.
        """
        return sorted(self.get_activation_map(wired_only=wired_only).keys())
    
    def forward(
        self, 
        tokens=None, 
        input_ids=None,
        return_type="logits",
        **kwargs
    ):
        """Forward pass through the model with TransformerLens-style interface."""
        # Handle input
        if tokens is not None:
            input_ids = tokens
        elif input_ids is None:
            raise ValueError("Must provide either 'tokens' or 'input_ids'")
            
        # Ensure tokens are on the right device
        if hasattr(input_ids, 'to'):
            input_ids = input_ids.to(next(self.predictor.parameters()).device)
        
        # Generate output based on return_type
        if return_type == "logits":
            return self.get_logits(input_ids=input_ids, **kwargs)
        elif return_type == "loss":
            logits = self.get_logits(input_ids=input_ids, **kwargs)
            targets = input_ids[:, 1:]
            logits = logits[:, :-1]
            loss = torch.nn.functional.cross_entropy(
                logits.reshape(-1, logits.size(-1)), 
                targets.reshape(-1),
                ignore_index=-100
            )
            return loss
        else:
            raise ValueError(f"Invalid return_type: {return_type}")
    
    def __call__(self, *args, **kwargs):
        """Make the model callable."""
        return self.forward(*args, **kwargs)
    
    # Core model functionality
    def get_logits(self, input_ids=None, text=None, **kwargs):
        """Get logits from input."""
        if text is not None:
            input_ids = self.to_tokens(text)
        
        # Ensure tokens are on the right device
        if hasattr(input_ids, 'to'):
            input_ids = input_ids.to(next(self.predictor.parameters()).device)
        
        outputs = self.predictor(input_ids, **kwargs)
        return outputs.logits
    
    def generate(self, text=None, input_ids=None, tokens=None, max_new_tokens=10, **kwargs):
        """Generate token ids (TransformerLens-compatible).

        Accepts `text` (str or list[str]) or pre-tokenized `tokens` / `input_ids` (Tensor).
        Returns the full generated token ids, including the prompt, on the model device.
        """
        # Normalize inputs
        if tokens is not None:
            input_ids = tokens
        elif input_ids is None and text is not None:
            input_ids = self.to_tokens(text)
        elif input_ids is None and isinstance(text, Tensor):
            input_ids = text

        if not isinstance(input_ids, Tensor):
            raise ValueError("generate expects `text` (str/list[str]) or `tokens`/`input_ids` (Tensor)")

        # Ensure tokens are on model device
        device = next(self.predictor.parameters()).device
        input_ids = input_ids.to(device)

        # Ensure a pad token id for HF generate
        pad_id = self.tokenizer.pad_token_id
        if pad_id is None:
            pad_id = self.tokenizer.eos_token_id
        
        with torch.no_grad():
            generated = self.predictor.generate(
                input_ids,
                max_new_tokens=max_new_tokens,
                pad_token_id=pad_id,
                **kwargs,
            )

        return generated

    def generate_text(self, *args, **kwargs) -> list[str]:
        """Convenience: decode generated tokens to strings."""
        generated = self.generate(*args, **kwargs)
        return self.tokenizer.batch_decode(generated, skip_special_tokens=True)
    
    # TransformerLens-style tokenizer utilities
    def to_tokens(self, text, prepend_bos=True):
        """Convert text (str or list[str]) to token ids Tensor [batch, seq]."""
        # If already a tensor, assume it's tokens and return as batch
        if isinstance(text, Tensor):
            return text if text.dim() == 2 else text.unsqueeze(0)

        if isinstance(text, str):
            text = [text]
        
        encoded_list = []
        for t in text:
            if prepend_bos and getattr(self.tokenizer, 'bos_token', None):
                t = (self.tokenizer.bos_token or "") + t
            encoded = self.tokenizer.encode(t, return_tensors="pt", add_special_tokens=False)
            encoded_list.append(encoded.squeeze(0))
        
        # Pad to same length
        if len(encoded_list) > 1:
            max_len = max(t.size(0) for t in encoded_list)
            padded = []
            for t in encoded_list:
                if t.size(0) < max_len:
                    pad_length = max_len - t.size(0)
                    pad_val = self.tokenizer.pad_token_id
                    if pad_val is None:
                        pad_val = self.tokenizer.eos_token_id
                    t = torch.cat([t, torch.full((pad_length,), pad_val, dtype=t.dtype)])
                padded.append(t)
            tokens = torch.stack(padded)
        else:
            tokens = encoded_list[0].unsqueeze(0)
        return tokens.to(self.cfg.device)
    
    def to_string(self, tokens):
        """Convert tokens to string(s).

        - If `tokens` is shape [seq], returns a single string.
        - If `tokens` is shape [batch, seq], returns a list[str] (one per batch element).
        - Also accepts list[int] or list[list[int]].
        """
        # Normalize to CPU and python lists
        if isinstance(tokens, Tensor):
            if tokens.dim() == 1:
                ids = tokens.detach().cpu().tolist()
                return self.tokenizer.decode(ids, skip_special_tokens=True)
            elif tokens.dim() == 2:
                batch_ids = [row.detach().cpu().tolist() for row in tokens]
                return self.tokenizer.batch_decode(batch_ids, skip_special_tokens=True)
            else:
                raise ValueError("tokens must be a 1D or 2D Tensor")

        # Handle python lists
        if isinstance(tokens, list):
            if len(tokens) == 0:
                return ""
            # list of list[int]
            if isinstance(tokens[0], list):
                return self.tokenizer.batch_decode(tokens, skip_special_tokens=True)
            # list[int]
            else:
                return self.tokenizer.decode(tokens, skip_special_tokens=True)

        raise ValueError("Unsupported tokens type for to_string")
    
    def to_str_tokens(self, tokens, pretty: bool = True, remove_special: bool = False):
        """Convert token ids to a list of string tokens (optionally pretty-print).

        pretty=True will map common BPE/SentencePiece markers (eg "Ġ" -> space, "Ċ" -> \n, "▁" -> space).
        remove_special=True will drop tokens that tokenizer flags as special tokens.
        """
        # Normalize to batch list of lists
        if isinstance(tokens, Tensor):
            if tokens.dim() == 1:
                token_id_seqs = [tokens.tolist()]
            elif tokens.dim() == 2:
                token_id_seqs = [row.tolist() for row in tokens]
            else:
                raise ValueError("tokens must be 1D or 2D Tensor")
        else:
            # Assume list of ids or list of list of ids
            if len(tokens) > 0 and isinstance(tokens[0], list):
                token_id_seqs = tokens
            else:
                token_id_seqs = [tokens]

        out: list[list[str]] = []
        for ids in token_id_seqs:
            raw_tokens = self.tokenizer.convert_ids_to_tokens(ids)

            if remove_special and hasattr(self.tokenizer, "all_special_tokens"):
                specials = set(self.tokenizer.all_special_tokens)
            else:
                specials = set()

            if not pretty:
                out.append([tok for tok in raw_tokens if tok not in specials])
                continue

            pretty_tokens: list[str] = []
            for i, tok in enumerate(raw_tokens):
                if tok in specials:
                    if not remove_special:
                        pretty_tokens.append(tok)
                    continue

                # GPT-2 style byte-level BPE markers
                if tok.startswith("Ġ"):
                    tok = " " + tok[1:]
                tok = tok.replace("Ċ", "\n")

                # SentencePiece marker for space
                if tok.startswith("▁"):
                    # add a space if not the first token
                    tok = (" " if len(pretty_tokens) > 0 else "") + tok.replace("▁", "", 1)

                pretty_tokens.append(tok)

            out.append(pretty_tokens)

        return out if len(out) > 1 else out[0]

    def to_single_token(self, string: str) -> int:
        """Map a string that is exactly one token to its token id.

        Raises an error if the input string does not correspond to a single token.
        """
        token = self.to_tokens(string, prepend_bos=False).squeeze()
        try: 
            assert not token.shape, f"Input string: {string} is not a single token!"
            return token.item()
        except AssertionError:
            print(f"WARNING: Input string: {string} is not a single token! Got {self.to_str_tokens(token)}")
            token = token[-1]
            return token.item()

    # Utility methods
    def to(self, device):
        """Move model to device."""
        self.cfg.device = device
        if hasattr(self, 'predictor') and self.predictor:
            self.predictor.to(device)
        return self

    def show_details(self, verbose=True):
        """Show model details."""
        if verbose:
            print(f" - n_layers : {self.cfg.n_layers}")
            print(f" - d_model  : {self.cfg.d_model}")
            print(f" - n_heads  : {self.cfg.n_heads}")
            print(f" - d_head   : {self.cfg.d_head}")
            print(f" - d_mlp    : {self.cfg.d_mlp}")
        else:
            print(f" - n_layers, d_model = {self.cfg.n_layers}, {self.cfg.d_model}")
